diff --git a/core/application/agent_jobs.go b/core/application/agent_jobs.go
new file mode 100644
index 000000000..b4e28aa5a
--- /dev/null
+++ b/core/application/agent_jobs.go
@@ -0,0 +1,43 @@
+package application
+
+import (
+ "time"
+
+ "github.com/mudler/LocalAI/core/services"
+ "github.com/rs/zerolog/log"
+)
+
+// RestartAgentJobService restarts the agent job service with current ApplicationConfig settings
+func (a *Application) RestartAgentJobService() error {
+ a.agentJobMutex.Lock()
+ defer a.agentJobMutex.Unlock()
+
+ // Stop existing service if running
+ if a.agentJobService != nil {
+ if err := a.agentJobService.Stop(); err != nil {
+ log.Warn().Err(err).Msg("Error stopping agent job service")
+ }
+ // Wait a bit for shutdown to complete
+ time.Sleep(200 * time.Millisecond)
+ }
+
+ // Create new service instance
+ agentJobService := services.NewAgentJobService(
+ a.ApplicationConfig(),
+ a.ModelLoader(),
+ a.ModelConfigLoader(),
+ a.TemplatesEvaluator(),
+ )
+
+ // Start the service
+ err := agentJobService.Start(a.ApplicationConfig().Context)
+ if err != nil {
+ log.Error().Err(err).Msg("Failed to start agent job service")
+ return err
+ }
+
+ a.agentJobService = agentJobService
+ log.Info().Msg("Agent job service restarted")
+ return nil
+}
+
diff --git a/core/application/application.go b/core/application/application.go
index 24c53fcba..3e241c698 100644
--- a/core/application/application.go
+++ b/core/application/application.go
@@ -17,11 +17,13 @@ type Application struct {
startupConfig *config.ApplicationConfig // Stores original config from env vars (before file loading)
templatesEvaluator *templates.Evaluator
galleryService *services.GalleryService
+ agentJobService *services.AgentJobService
watchdogMutex sync.Mutex
watchdogStop chan bool
p2pMutex sync.Mutex
p2pCtx context.Context
p2pCancel context.CancelFunc
+ agentJobMutex sync.Mutex
}
func newApplication(appConfig *config.ApplicationConfig) *Application {
@@ -53,6 +55,10 @@ func (a *Application) GalleryService() *services.GalleryService {
return a.galleryService
}
+func (a *Application) AgentJobService() *services.AgentJobService {
+ return a.agentJobService
+}
+
// StartupConfig returns the original startup configuration (from env vars, before file loading)
func (a *Application) StartupConfig() *config.ApplicationConfig {
return a.startupConfig
@@ -67,5 +73,20 @@ func (a *Application) start() error {
a.galleryService = galleryService
+ // Initialize agent job service
+ agentJobService := services.NewAgentJobService(
+ a.ApplicationConfig(),
+ a.ModelLoader(),
+ a.ModelConfigLoader(),
+ a.TemplatesEvaluator(),
+ )
+
+ err = agentJobService.Start(a.ApplicationConfig().Context)
+ if err != nil {
+ return err
+ }
+
+ a.agentJobService = agentJobService
+
return nil
}
diff --git a/core/application/config_file_watcher.go b/core/application/config_file_watcher.go
index 0129828ca..999d29aec 100644
--- a/core/application/config_file_watcher.go
+++ b/core/application/config_file_watcher.go
@@ -43,6 +43,8 @@ func newConfigFileHandler(appConfig *config.ApplicationConfig) configFileHandler
if err != nil {
log.Error().Err(err).Str("file", "runtime_settings.json").Msg("unable to register config file handler")
}
+ // Note: agent_tasks.json and agent_jobs.json are handled by AgentJobService directly
+ // The service watches and reloads these files internally
return c
}
@@ -206,6 +208,7 @@ type runtimeSettings struct {
AutoloadGalleries *bool `json:"autoload_galleries,omitempty"`
AutoloadBackendGalleries *bool `json:"autoload_backend_galleries,omitempty"`
ApiKeys *[]string `json:"api_keys,omitempty"`
+ AgentJobRetentionDays *int `json:"agent_job_retention_days,omitempty"`
}
func readRuntimeSettingsJson(startupAppConfig config.ApplicationConfig) fileHandler {
@@ -234,6 +237,7 @@ func readRuntimeSettingsJson(startupAppConfig config.ApplicationConfig) fileHand
envFederated := appConfig.Federated == startupAppConfig.Federated
envAutoloadGalleries := appConfig.AutoloadGalleries == startupAppConfig.AutoloadGalleries
envAutoloadBackendGalleries := appConfig.AutoloadBackendGalleries == startupAppConfig.AutoloadBackendGalleries
+ envAgentJobRetentionDays := appConfig.AgentJobRetentionDays == startupAppConfig.AgentJobRetentionDays
if len(fileContent) > 0 {
var settings runtimeSettings
@@ -328,6 +332,9 @@ func readRuntimeSettingsJson(startupAppConfig config.ApplicationConfig) fileHand
// Replace all runtime keys with what's in runtime_settings.json
appConfig.ApiKeys = append(envKeys, runtimeKeys...)
}
+ if settings.AgentJobRetentionDays != nil && !envAgentJobRetentionDays {
+ appConfig.AgentJobRetentionDays = *settings.AgentJobRetentionDays
+ }
// If watchdog is enabled via file but not via env, ensure WatchDog flag is set
if !envWatchdogIdle && !envWatchdogBusy {
diff --git a/core/application/startup.go b/core/application/startup.go
index 6186424e5..2bbbdfac7 100644
--- a/core/application/startup.go
+++ b/core/application/startup.go
@@ -226,6 +226,7 @@ func loadRuntimeSettingsFromFile(options *config.ApplicationConfig) {
WatchdogBusyTimeout *string `json:"watchdog_busy_timeout,omitempty"`
SingleBackend *bool `json:"single_backend,omitempty"`
ParallelBackendRequests *bool `json:"parallel_backend_requests,omitempty"`
+ AgentJobRetentionDays *int `json:"agent_job_retention_days,omitempty"`
}
if err := json.Unmarshal(fileContent, &settings); err != nil {
@@ -289,6 +290,12 @@ func loadRuntimeSettingsFromFile(options *config.ApplicationConfig) {
options.ParallelBackendRequests = *settings.ParallelBackendRequests
}
}
+ if settings.AgentJobRetentionDays != nil {
+ // Only apply if current value is default (0), suggesting it wasn't set from env var
+ if options.AgentJobRetentionDays == 0 {
+ options.AgentJobRetentionDays = *settings.AgentJobRetentionDays
+ }
+ }
if !options.WatchDogIdle && !options.WatchDogBusy {
if settings.WatchdogEnabled != nil && *settings.WatchdogEnabled {
options.WatchDog = true
diff --git a/core/cli/run.go b/core/cli/run.go
index a1dc0e1c1..3cc77baf1 100644
--- a/core/cli/run.go
+++ b/core/cli/run.go
@@ -75,6 +75,7 @@ type RunCMD struct {
DisableGalleryEndpoint bool `env:"LOCALAI_DISABLE_GALLERY_ENDPOINT,DISABLE_GALLERY_ENDPOINT" help:"Disable the gallery endpoints" group:"api"`
MachineTag string `env:"LOCALAI_MACHINE_TAG,MACHINE_TAG" help:"Add Machine-Tag header to each response which is useful to track the machine in the P2P network" group:"api"`
LoadToMemory []string `env:"LOCALAI_LOAD_TO_MEMORY,LOAD_TO_MEMORY" help:"A list of models to load into memory at startup" group:"models"`
+ AgentJobRetentionDays int `env:"LOCALAI_AGENT_JOB_RETENTION_DAYS,AGENT_JOB_RETENTION_DAYS" default:"30" help:"Number of days to keep agent job history (default: 30)" group:"api"`
Version bool
}
@@ -129,6 +130,7 @@ func (r *RunCMD) Run(ctx *cliContext.Context) error {
config.WithLoadToMemory(r.LoadToMemory),
config.WithMachineTag(r.MachineTag),
config.WithAPIAddress(r.Address),
+ config.WithAgentJobRetentionDays(r.AgentJobRetentionDays),
config.WithTunnelCallback(func(tunnels []string) {
tunnelEnvVar := strings.Join(tunnels, ",")
// TODO: this is very specific to llama.cpp, we should have a more generic way to set the environment variable
diff --git a/core/config/application_config.go b/core/config/application_config.go
index 9a9a8171c..6f94a04a4 100644
--- a/core/config/application_config.go
+++ b/core/config/application_config.go
@@ -70,15 +70,18 @@ type ApplicationConfig struct {
TunnelCallback func(tunnels []string)
DisableRuntimeSettings bool
+
+ AgentJobRetentionDays int // Default: 30 days
}
type AppOption func(*ApplicationConfig)
func NewApplicationConfig(o ...AppOption) *ApplicationConfig {
opt := &ApplicationConfig{
- Context: context.Background(),
- UploadLimitMB: 15,
- Debug: true,
+ Context: context.Background(),
+ UploadLimitMB: 15,
+ Debug: true,
+ AgentJobRetentionDays: 30, // Default: 30 days
}
for _, oo := range o {
oo(opt)
@@ -333,6 +336,12 @@ func WithApiKeys(apiKeys []string) AppOption {
}
}
+func WithAgentJobRetentionDays(days int) AppOption {
+ return func(o *ApplicationConfig) {
+ o.AgentJobRetentionDays = days
+ }
+}
+
func WithEnforcedPredownloadScans(enforced bool) AppOption {
return func(o *ApplicationConfig) {
o.EnforcePredownloadScans = enforced
diff --git a/core/http/app.go b/core/http/app.go
index a5ce91e42..b12f66c31 100644
--- a/core/http/app.go
+++ b/core/http/app.go
@@ -205,7 +205,7 @@ func API(application *application.Application) (*echo.Echo, error) {
opcache = services.NewOpCache(application.GalleryService())
}
- routes.RegisterLocalAIRoutes(e, requestExtractor, application.ModelConfigLoader(), application.ModelLoader(), application.ApplicationConfig(), application.GalleryService(), opcache, application.TemplatesEvaluator())
+ routes.RegisterLocalAIRoutes(e, requestExtractor, application.ModelConfigLoader(), application.ModelLoader(), application.ApplicationConfig(), application.GalleryService(), opcache, application.TemplatesEvaluator(), application)
routes.RegisterOpenAIRoutes(e, requestExtractor, application)
if !application.ApplicationConfig().DisableWebUI {
routes.RegisterUIAPIRoutes(e, application.ModelConfigLoader(), application.ModelLoader(), application.ApplicationConfig(), application.GalleryService(), opcache, application)
diff --git a/core/http/app_test.go b/core/http/app_test.go
index 362a6bc69..a733fcdd6 100644
--- a/core/http/app_test.go
+++ b/core/http/app_test.go
@@ -210,6 +210,41 @@ func postRequestResponseJSON[B1 any, B2 any](url string, reqJson *B1, respJson *
return json.Unmarshal(body, respJson)
}
+func putRequestJSON[B any](url string, bodyJson *B) error {
+ payload, err := json.Marshal(bodyJson)
+ if err != nil {
+ return err
+ }
+
+ GinkgoWriter.Printf("PUT %s: %s\n", url, string(payload))
+
+ req, err := http.NewRequest("PUT", url, bytes.NewBuffer(payload))
+ if err != nil {
+ return err
+ }
+
+ req.Header.Set("Content-Type", "application/json")
+ req.Header.Set("Authorization", bearerKey)
+
+ client := &http.Client{}
+ resp, err := client.Do(req)
+ if err != nil {
+ return err
+ }
+ defer resp.Body.Close()
+
+ body, err := io.ReadAll(resp.Body)
+ if err != nil {
+ return err
+ }
+
+ if resp.StatusCode < 200 || resp.StatusCode >= 400 {
+ return fmt.Errorf("unexpected status code: %d, body: %s", resp.StatusCode, string(body))
+ }
+
+ return nil
+}
+
func postInvalidRequest(url string) (error, int) {
req, err := http.NewRequest("POST", url, bytes.NewBufferString("invalid request"))
@@ -1194,6 +1229,138 @@ parameters:
Expect(findRespBody.Similarities[i]).To(BeNumerically("<=", 1))
}
})
+
+ Context("Agent Jobs", Label("agent-jobs"), func() {
+ It("creates and manages tasks", func() {
+ // Create a task
+ taskBody := map[string]interface{}{
+ "name": "Test Task",
+ "description": "Test Description",
+ "model": "testmodel.ggml",
+ "prompt": "Hello {{.name}}",
+ "enabled": true,
+ }
+
+ var createResp map[string]interface{}
+ err := postRequestResponseJSON("http://127.0.0.1:9090/api/agent/tasks", &taskBody, &createResp)
+ Expect(err).ToNot(HaveOccurred())
+ Expect(createResp["id"]).ToNot(BeEmpty())
+ taskID := createResp["id"].(string)
+
+ // Get the task
+ var task schema.Task
+ resp, err := http.Get("http://127.0.0.1:9090/api/agent/tasks/" + taskID)
+ Expect(err).ToNot(HaveOccurred())
+ Expect(resp.StatusCode).To(Equal(200))
+ body, _ := io.ReadAll(resp.Body)
+ json.Unmarshal(body, &task)
+ Expect(task.Name).To(Equal("Test Task"))
+
+ // List tasks
+ resp, err = http.Get("http://127.0.0.1:9090/api/agent/tasks")
+ Expect(err).ToNot(HaveOccurred())
+ Expect(resp.StatusCode).To(Equal(200))
+ var tasks []schema.Task
+ body, _ = io.ReadAll(resp.Body)
+ json.Unmarshal(body, &tasks)
+ Expect(len(tasks)).To(BeNumerically(">=", 1))
+
+ // Update task
+ taskBody["name"] = "Updated Task"
+ err = putRequestJSON("http://127.0.0.1:9090/api/agent/tasks/"+taskID, &taskBody)
+ Expect(err).ToNot(HaveOccurred())
+
+ // Verify update
+ resp, err = http.Get("http://127.0.0.1:9090/api/agent/tasks/" + taskID)
+ Expect(err).ToNot(HaveOccurred())
+ body, _ = io.ReadAll(resp.Body)
+ json.Unmarshal(body, &task)
+ Expect(task.Name).To(Equal("Updated Task"))
+
+ // Delete task
+ req, _ := http.NewRequest("DELETE", "http://127.0.0.1:9090/api/agent/tasks/"+taskID, nil)
+ req.Header.Set("Authorization", bearerKey)
+ resp, err = http.DefaultClient.Do(req)
+ Expect(err).ToNot(HaveOccurred())
+ Expect(resp.StatusCode).To(Equal(200))
+ })
+
+ It("executes and monitors jobs", func() {
+ // Create a task first
+ taskBody := map[string]interface{}{
+ "name": "Job Test Task",
+ "model": "testmodel.ggml",
+ "prompt": "Say hello",
+ "enabled": true,
+ }
+
+ var createResp map[string]interface{}
+ err := postRequestResponseJSON("http://127.0.0.1:9090/api/agent/tasks", &taskBody, &createResp)
+ Expect(err).ToNot(HaveOccurred())
+ taskID := createResp["id"].(string)
+
+ // Execute a job
+ jobBody := map[string]interface{}{
+ "task_id": taskID,
+ "parameters": map[string]string{},
+ }
+
+ var jobResp schema.JobExecutionResponse
+ err = postRequestResponseJSON("http://127.0.0.1:9090/api/agent/jobs/execute", &jobBody, &jobResp)
+ Expect(err).ToNot(HaveOccurred())
+ Expect(jobResp.JobID).ToNot(BeEmpty())
+ jobID := jobResp.JobID
+
+ // Get job status
+ var job schema.Job
+ resp, err := http.Get("http://127.0.0.1:9090/api/agent/jobs/" + jobID)
+ Expect(err).ToNot(HaveOccurred())
+ Expect(resp.StatusCode).To(Equal(200))
+ body, _ := io.ReadAll(resp.Body)
+ json.Unmarshal(body, &job)
+ Expect(job.ID).To(Equal(jobID))
+ Expect(job.TaskID).To(Equal(taskID))
+
+ // List jobs
+ resp, err = http.Get("http://127.0.0.1:9090/api/agent/jobs")
+ Expect(err).ToNot(HaveOccurred())
+ Expect(resp.StatusCode).To(Equal(200))
+ var jobs []schema.Job
+ body, _ = io.ReadAll(resp.Body)
+ json.Unmarshal(body, &jobs)
+ Expect(len(jobs)).To(BeNumerically(">=", 1))
+
+ // Cancel job (if still pending/running)
+ if job.Status == schema.JobStatusPending || job.Status == schema.JobStatusRunning {
+ req, _ := http.NewRequest("POST", "http://127.0.0.1:9090/api/agent/jobs/"+jobID+"/cancel", nil)
+ req.Header.Set("Authorization", bearerKey)
+ resp, err = http.DefaultClient.Do(req)
+ Expect(err).ToNot(HaveOccurred())
+ Expect(resp.StatusCode).To(Equal(200))
+ }
+ })
+
+ It("executes task by name", func() {
+ // Create a task with a specific name
+ taskBody := map[string]interface{}{
+ "name": "Named Task",
+ "model": "testmodel.ggml",
+ "prompt": "Hello",
+ "enabled": true,
+ }
+
+ var createResp map[string]interface{}
+ err := postRequestResponseJSON("http://127.0.0.1:9090/api/agent/tasks", &taskBody, &createResp)
+ Expect(err).ToNot(HaveOccurred())
+
+ // Execute by name
+ paramsBody := map[string]string{"param1": "value1"}
+ var jobResp schema.JobExecutionResponse
+ err = postRequestResponseJSON("http://127.0.0.1:9090/api/agent/tasks/Named Task/execute", ¶msBody, &jobResp)
+ Expect(err).ToNot(HaveOccurred())
+ Expect(jobResp.JobID).ToNot(BeEmpty())
+ })
+ })
})
})
diff --git a/core/http/endpoints/localai/agent_jobs.go b/core/http/endpoints/localai/agent_jobs.go
new file mode 100644
index 000000000..0e65a241d
--- /dev/null
+++ b/core/http/endpoints/localai/agent_jobs.go
@@ -0,0 +1,339 @@
+package localai
+
+import (
+ "fmt"
+ "net/http"
+ "strconv"
+
+ "github.com/labstack/echo/v4"
+ "github.com/mudler/LocalAI/core/application"
+ "github.com/mudler/LocalAI/core/schema"
+)
+
+// CreateTaskEndpoint creates a new agent task
+// @Summary Create a new agent task
+// @Description Create a new reusable agent task with prompt template and configuration
+// @Tags agent-jobs
+// @Accept json
+// @Produce json
+// @Param task body schema.Task true "Task definition"
+// @Success 201 {object} map[string]string "Task created"
+// @Failure 400 {object} map[string]string "Invalid request"
+// @Failure 500 {object} map[string]string "Internal server error"
+// @Router /api/agent/tasks [post]
+func CreateTaskEndpoint(app *application.Application) echo.HandlerFunc {
+ return func(c echo.Context) error {
+ var task schema.Task
+ if err := c.Bind(&task); err != nil {
+ return c.JSON(http.StatusBadRequest, map[string]string{"error": "Invalid request body: " + err.Error()})
+ }
+
+ id, err := app.AgentJobService().CreateTask(task)
+ if err != nil {
+ return c.JSON(http.StatusBadRequest, map[string]string{"error": err.Error()})
+ }
+
+ return c.JSON(http.StatusCreated, map[string]string{"id": id})
+ }
+}
+
+// UpdateTaskEndpoint updates an existing task
+// @Summary Update an agent task
+// @Description Update an existing agent task
+// @Tags agent-jobs
+// @Accept json
+// @Produce json
+// @Param id path string true "Task ID"
+// @Param task body schema.Task true "Updated task definition"
+// @Success 200 {object} map[string]string "Task updated"
+// @Failure 400 {object} map[string]string "Invalid request"
+// @Failure 404 {object} map[string]string "Task not found"
+// @Router /api/agent/tasks/{id} [put]
+func UpdateTaskEndpoint(app *application.Application) echo.HandlerFunc {
+ return func(c echo.Context) error {
+ id := c.Param("id")
+ var task schema.Task
+ if err := c.Bind(&task); err != nil {
+ return c.JSON(http.StatusBadRequest, map[string]string{"error": "Invalid request body: " + err.Error()})
+ }
+
+ if err := app.AgentJobService().UpdateTask(id, task); err != nil {
+ if err.Error() == "task not found: "+id {
+ return c.JSON(http.StatusNotFound, map[string]string{"error": err.Error()})
+ }
+ return c.JSON(http.StatusBadRequest, map[string]string{"error": err.Error()})
+ }
+
+ return c.JSON(http.StatusOK, map[string]string{"message": "Task updated"})
+ }
+}
+
+// DeleteTaskEndpoint deletes a task
+// @Summary Delete an agent task
+// @Description Delete an agent task by ID
+// @Tags agent-jobs
+// @Produce json
+// @Param id path string true "Task ID"
+// @Success 200 {object} map[string]string "Task deleted"
+// @Failure 404 {object} map[string]string "Task not found"
+// @Router /api/agent/tasks/{id} [delete]
+func DeleteTaskEndpoint(app *application.Application) echo.HandlerFunc {
+ return func(c echo.Context) error {
+ id := c.Param("id")
+ if err := app.AgentJobService().DeleteTask(id); err != nil {
+ if err.Error() == "task not found: "+id {
+ return c.JSON(http.StatusNotFound, map[string]string{"error": err.Error()})
+ }
+ return c.JSON(http.StatusInternalServerError, map[string]string{"error": err.Error()})
+ }
+
+ return c.JSON(http.StatusOK, map[string]string{"message": "Task deleted"})
+ }
+}
+
+// ListTasksEndpoint lists all tasks
+// @Summary List all agent tasks
+// @Description Get a list of all agent tasks
+// @Tags agent-jobs
+// @Produce json
+// @Success 200 {array} schema.Task "List of tasks"
+// @Router /api/agent/tasks [get]
+func ListTasksEndpoint(app *application.Application) echo.HandlerFunc {
+ return func(c echo.Context) error {
+ tasks := app.AgentJobService().ListTasks()
+ return c.JSON(http.StatusOK, tasks)
+ }
+}
+
+// GetTaskEndpoint gets a task by ID
+// @Summary Get an agent task
+// @Description Get an agent task by ID
+// @Tags agent-jobs
+// @Produce json
+// @Param id path string true "Task ID"
+// @Success 200 {object} schema.Task "Task details"
+// @Failure 404 {object} map[string]string "Task not found"
+// @Router /api/agent/tasks/{id} [get]
+func GetTaskEndpoint(app *application.Application) echo.HandlerFunc {
+ return func(c echo.Context) error {
+ id := c.Param("id")
+ task, err := app.AgentJobService().GetTask(id)
+ if err != nil {
+ return c.JSON(http.StatusNotFound, map[string]string{"error": err.Error()})
+ }
+
+ return c.JSON(http.StatusOK, task)
+ }
+}
+
+// ExecuteJobEndpoint executes a job
+// @Summary Execute an agent job
+// @Description Create and execute a new agent job
+// @Tags agent-jobs
+// @Accept json
+// @Produce json
+// @Param request body schema.JobExecutionRequest true "Job execution request"
+// @Success 201 {object} schema.JobExecutionResponse "Job created"
+// @Failure 400 {object} map[string]string "Invalid request"
+// @Router /api/agent/jobs/execute [post]
+func ExecuteJobEndpoint(app *application.Application) echo.HandlerFunc {
+ return func(c echo.Context) error {
+ var req schema.JobExecutionRequest
+ if err := c.Bind(&req); err != nil {
+ return c.JSON(http.StatusBadRequest, map[string]string{"error": "Invalid request body: " + err.Error()})
+ }
+
+ if req.Parameters == nil {
+ req.Parameters = make(map[string]string)
+ }
+
+ jobID, err := app.AgentJobService().ExecuteJob(req.TaskID, req.Parameters, "api")
+ if err != nil {
+ return c.JSON(http.StatusBadRequest, map[string]string{"error": err.Error()})
+ }
+
+ baseURL := c.Scheme() + "://" + c.Request().Host
+ return c.JSON(http.StatusCreated, schema.JobExecutionResponse{
+ JobID: jobID,
+ Status: "pending",
+ URL: baseURL + "/api/agent/jobs/" + jobID,
+ })
+ }
+}
+
+// GetJobEndpoint gets a job by ID
+// @Summary Get an agent job
+// @Description Get an agent job by ID
+// @Tags agent-jobs
+// @Produce json
+// @Param id path string true "Job ID"
+// @Success 200 {object} schema.Job "Job details"
+// @Failure 404 {object} map[string]string "Job not found"
+// @Router /api/agent/jobs/{id} [get]
+func GetJobEndpoint(app *application.Application) echo.HandlerFunc {
+ return func(c echo.Context) error {
+ id := c.Param("id")
+ job, err := app.AgentJobService().GetJob(id)
+ if err != nil {
+ return c.JSON(http.StatusNotFound, map[string]string{"error": err.Error()})
+ }
+
+ return c.JSON(http.StatusOK, job)
+ }
+}
+
+// ListJobsEndpoint lists jobs with optional filtering
+// @Summary List agent jobs
+// @Description Get a list of agent jobs, optionally filtered by task_id and status
+// @Tags agent-jobs
+// @Produce json
+// @Param task_id query string false "Filter by task ID"
+// @Param status query string false "Filter by status (pending, running, completed, failed, cancelled)"
+// @Param limit query int false "Limit number of results"
+// @Success 200 {array} schema.Job "List of jobs"
+// @Router /api/agent/jobs [get]
+func ListJobsEndpoint(app *application.Application) echo.HandlerFunc {
+ return func(c echo.Context) error {
+ var taskID *string
+ var status *schema.JobStatus
+ limit := 0
+
+ if taskIDParam := c.QueryParam("task_id"); taskIDParam != "" {
+ taskID = &taskIDParam
+ }
+
+ if statusParam := c.QueryParam("status"); statusParam != "" {
+ s := schema.JobStatus(statusParam)
+ status = &s
+ }
+
+ if limitParam := c.QueryParam("limit"); limitParam != "" {
+ if l, err := strconv.Atoi(limitParam); err == nil {
+ limit = l
+ }
+ }
+
+ jobs := app.AgentJobService().ListJobs(taskID, status, limit)
+ return c.JSON(http.StatusOK, jobs)
+ }
+}
+
+// CancelJobEndpoint cancels a running job
+// @Summary Cancel an agent job
+// @Description Cancel a running or pending agent job
+// @Tags agent-jobs
+// @Produce json
+// @Param id path string true "Job ID"
+// @Success 200 {object} map[string]string "Job cancelled"
+// @Failure 400 {object} map[string]string "Job cannot be cancelled"
+// @Failure 404 {object} map[string]string "Job not found"
+// @Router /api/agent/jobs/{id}/cancel [post]
+func CancelJobEndpoint(app *application.Application) echo.HandlerFunc {
+ return func(c echo.Context) error {
+ id := c.Param("id")
+ if err := app.AgentJobService().CancelJob(id); err != nil {
+ if err.Error() == "job not found: "+id {
+ return c.JSON(http.StatusNotFound, map[string]string{"error": err.Error()})
+ }
+ return c.JSON(http.StatusBadRequest, map[string]string{"error": err.Error()})
+ }
+
+ return c.JSON(http.StatusOK, map[string]string{"message": "Job cancelled"})
+ }
+}
+
+// DeleteJobEndpoint deletes a job
+// @Summary Delete an agent job
+// @Description Delete an agent job by ID
+// @Tags agent-jobs
+// @Produce json
+// @Param id path string true "Job ID"
+// @Success 200 {object} map[string]string "Job deleted"
+// @Failure 404 {object} map[string]string "Job not found"
+// @Router /api/agent/jobs/{id} [delete]
+func DeleteJobEndpoint(app *application.Application) echo.HandlerFunc {
+ return func(c echo.Context) error {
+ id := c.Param("id")
+ if err := app.AgentJobService().DeleteJob(id); err != nil {
+ if err.Error() == "job not found: "+id {
+ return c.JSON(http.StatusNotFound, map[string]string{"error": err.Error()})
+ }
+ return c.JSON(http.StatusInternalServerError, map[string]string{"error": err.Error()})
+ }
+
+ return c.JSON(http.StatusOK, map[string]string{"message": "Job deleted"})
+ }
+}
+
+// ExecuteTaskByNameEndpoint executes a task by name
+// @Summary Execute a task by name
+// @Description Execute an agent task by its name (convenience endpoint). Parameters can be provided in the request body as a JSON object with string values.
+// @Tags agent-jobs
+// @Accept json
+// @Produce json
+// @Param name path string true "Task name"
+// @Param request body map[string]string false "Template parameters (JSON object with string values)"
+// @Success 201 {object} schema.JobExecutionResponse "Job created"
+// @Failure 400 {object} map[string]string "Invalid request"
+// @Failure 404 {object} map[string]string "Task not found"
+// @Router /api/agent/tasks/{name}/execute [post]
+func ExecuteTaskByNameEndpoint(app *application.Application) echo.HandlerFunc {
+ return func(c echo.Context) error {
+ name := c.Param("name")
+ var params map[string]string
+
+ // Try to bind parameters from request body
+ // If body is empty or invalid, use empty params
+ if c.Request().ContentLength > 0 {
+ if err := c.Bind(¶ms); err != nil {
+ // If binding fails, try to read as raw JSON
+ body := make(map[string]interface{})
+ if err := c.Bind(&body); err == nil {
+ // Convert interface{} values to strings
+ params = make(map[string]string)
+ for k, v := range body {
+ if str, ok := v.(string); ok {
+ params[k] = str
+ } else {
+ // Convert non-string values to string
+ params[k] = fmt.Sprintf("%v", v)
+ }
+ }
+ } else {
+ // If all binding fails, use empty params
+ params = make(map[string]string)
+ }
+ }
+ } else {
+ // No body provided, use empty params
+ params = make(map[string]string)
+ }
+
+ // Find task by name
+ tasks := app.AgentJobService().ListTasks()
+ var task *schema.Task
+ for _, t := range tasks {
+ if t.Name == name {
+ task = &t
+ break
+ }
+ }
+
+ if task == nil {
+ return c.JSON(http.StatusNotFound, map[string]string{"error": "Task not found: " + name})
+ }
+
+ jobID, err := app.AgentJobService().ExecuteJob(task.ID, params, "api")
+ if err != nil {
+ return c.JSON(http.StatusBadRequest, map[string]string{"error": err.Error()})
+ }
+
+ baseURL := c.Scheme() + "://" + c.Request().Host
+ return c.JSON(http.StatusCreated, schema.JobExecutionResponse{
+ JobID: jobID,
+ Status: "pending",
+ URL: baseURL + "/api/agent/jobs/" + jobID,
+ })
+ }
+}
+
diff --git a/core/http/endpoints/localai/settings.go b/core/http/endpoints/localai/settings.go
index 62f198a9d..d5c5cd7db 100644
--- a/core/http/endpoints/localai/settings.go
+++ b/core/http/endpoints/localai/settings.go
@@ -44,6 +44,7 @@ type RuntimeSettings struct {
AutoloadGalleries *bool `json:"autoload_galleries,omitempty"`
AutoloadBackendGalleries *bool `json:"autoload_backend_galleries,omitempty"`
ApiKeys *[]string `json:"api_keys"` // No omitempty - we need to save empty arrays to clear keys
+ AgentJobRetentionDays *int `json:"agent_job_retention_days,omitempty"`
}
// GetSettingsEndpoint returns current settings with precedence (env > file > defaults)
@@ -80,6 +81,7 @@ func GetSettingsEndpoint(app *application.Application) echo.HandlerFunc {
autoloadGalleries := appConfig.AutoloadGalleries
autoloadBackendGalleries := appConfig.AutoloadBackendGalleries
apiKeys := appConfig.ApiKeys
+ agentJobRetentionDays := appConfig.AgentJobRetentionDays
settings.WatchdogIdleEnabled = &watchdogIdle
settings.WatchdogBusyEnabled = &watchdogBusy
@@ -101,6 +103,7 @@ func GetSettingsEndpoint(app *application.Application) echo.HandlerFunc {
settings.AutoloadGalleries = &autoloadGalleries
settings.AutoloadBackendGalleries = &autoloadBackendGalleries
settings.ApiKeys = &apiKeys
+ settings.AgentJobRetentionDays = &agentJobRetentionDays
var idleTimeout, busyTimeout string
if appConfig.WatchDogIdleTimeout > 0 {
@@ -268,6 +271,11 @@ func UpdateSettingsEndpoint(app *application.Application) echo.HandlerFunc {
if settings.AutoloadBackendGalleries != nil {
appConfig.AutoloadBackendGalleries = *settings.AutoloadBackendGalleries
}
+ agentJobChanged := false
+ if settings.AgentJobRetentionDays != nil {
+ appConfig.AgentJobRetentionDays = *settings.AgentJobRetentionDays
+ agentJobChanged = true
+ }
if settings.ApiKeys != nil {
// API keys from env vars (startup) should be kept, runtime settings keys are added
// Combine startup keys (env vars) with runtime settings keys
@@ -302,6 +310,17 @@ func UpdateSettingsEndpoint(app *application.Application) echo.HandlerFunc {
}
}
+ // Restart agent job service if retention days changed
+ if agentJobChanged {
+ if err := app.RestartAgentJobService(); err != nil {
+ log.Error().Err(err).Msg("Failed to restart agent job service")
+ return c.JSON(http.StatusInternalServerError, SettingsResponse{
+ Success: false,
+ Error: "Settings saved but failed to restart agent job service: " + err.Error(),
+ })
+ }
+ }
+
// Restart P2P if P2P settings changed
p2pChanged := settings.P2PToken != nil || settings.P2PNetworkID != nil || settings.Federated != nil
if p2pChanged {
diff --git a/core/http/routes/localai.go b/core/http/routes/localai.go
index bf8a7bfb8..32e030bf3 100644
--- a/core/http/routes/localai.go
+++ b/core/http/routes/localai.go
@@ -2,6 +2,7 @@ package routes
import (
"github.com/labstack/echo/v4"
+ "github.com/mudler/LocalAI/core/application"
"github.com/mudler/LocalAI/core/config"
"github.com/mudler/LocalAI/core/http/endpoints/localai"
"github.com/mudler/LocalAI/core/http/middleware"
@@ -20,7 +21,8 @@ func RegisterLocalAIRoutes(router *echo.Echo,
appConfig *config.ApplicationConfig,
galleryService *services.GalleryService,
opcache *services.OpCache,
- evaluator *templates.Evaluator) {
+ evaluator *templates.Evaluator,
+ app *application.Application) {
router.GET("/swagger/*", echoswagger.WrapHandler) // default
@@ -154,4 +156,21 @@ func RegisterLocalAIRoutes(router *echo.Echo,
router.POST("/mcp/v1/chat/completions", mcpStreamHandler, mcpStreamMiddleware...)
}
+ // Agent job routes
+ if app != nil && app.AgentJobService() != nil {
+ router.POST("/api/agent/tasks", localai.CreateTaskEndpoint(app))
+ router.PUT("/api/agent/tasks/:id", localai.UpdateTaskEndpoint(app))
+ router.DELETE("/api/agent/tasks/:id", localai.DeleteTaskEndpoint(app))
+ router.GET("/api/agent/tasks", localai.ListTasksEndpoint(app))
+ router.GET("/api/agent/tasks/:id", localai.GetTaskEndpoint(app))
+
+ router.POST("/api/agent/jobs/execute", localai.ExecuteJobEndpoint(app))
+ router.GET("/api/agent/jobs/:id", localai.GetJobEndpoint(app))
+ router.GET("/api/agent/jobs", localai.ListJobsEndpoint(app))
+ router.POST("/api/agent/jobs/:id/cancel", localai.CancelJobEndpoint(app))
+ router.DELETE("/api/agent/jobs/:id", localai.DeleteJobEndpoint(app))
+
+ router.POST("/api/agent/tasks/:name/execute", localai.ExecuteTaskByNameEndpoint(app))
+ }
+
}
diff --git a/core/http/routes/ui.go b/core/http/routes/ui.go
index 6ef565505..368aa4798 100644
--- a/core/http/routes/ui.go
+++ b/core/http/routes/ui.go
@@ -34,6 +34,60 @@ func RegisterUIRoutes(app *echo.Echo,
})
}
+ // Agent Jobs pages
+ app.GET("/agent-jobs", func(c echo.Context) error {
+ modelConfigs := cl.GetAllModelsConfigs()
+ summary := map[string]interface{}{
+ "Title": "LocalAI - Agent Jobs",
+ "BaseURL": middleware.BaseURL(c),
+ "Version": internal.PrintableVersion(),
+ "ModelsConfig": modelConfigs,
+ }
+ return c.Render(200, "views/agent-jobs", summary)
+ })
+
+ app.GET("/agent-jobs/tasks/new", func(c echo.Context) error {
+ modelConfigs := cl.GetAllModelsConfigs()
+ summary := map[string]interface{}{
+ "Title": "LocalAI - Create Task",
+ "BaseURL": middleware.BaseURL(c),
+ "Version": internal.PrintableVersion(),
+ "ModelsConfig": modelConfigs,
+ }
+ return c.Render(200, "views/agent-task-details", summary)
+ })
+
+ // More specific route must come first
+ app.GET("/agent-jobs/tasks/:id/edit", func(c echo.Context) error {
+ modelConfigs := cl.GetAllModelsConfigs()
+ summary := map[string]interface{}{
+ "Title": "LocalAI - Edit Task",
+ "BaseURL": middleware.BaseURL(c),
+ "Version": internal.PrintableVersion(),
+ "ModelsConfig": modelConfigs,
+ }
+ return c.Render(200, "views/agent-task-details", summary)
+ })
+
+ // Task details page (less specific, comes after edit route)
+ app.GET("/agent-jobs/tasks/:id", func(c echo.Context) error {
+ summary := map[string]interface{}{
+ "Title": "LocalAI - Task Details",
+ "BaseURL": middleware.BaseURL(c),
+ "Version": internal.PrintableVersion(),
+ }
+ return c.Render(200, "views/agent-task-details", summary)
+ })
+
+ app.GET("/agent-jobs/jobs/:id", func(c echo.Context) error {
+ summary := map[string]interface{}{
+ "Title": "LocalAI - Job Details",
+ "BaseURL": middleware.BaseURL(c),
+ "Version": internal.PrintableVersion(),
+ }
+ return c.Render(200, "views/agent-job-details", summary)
+ })
+
// P2P
app.GET("/p2p/", func(c echo.Context) error {
summary := map[string]interface{}{
diff --git a/core/http/views/agent-job-details.html b/core/http/views/agent-job-details.html
new file mode 100644
index 000000000..914da8019
--- /dev/null
+++ b/core/http/views/agent-job-details.html
@@ -0,0 +1,327 @@
+
+
+{{template "views/partials/head" .}}
+
+
+
+
+ {{template "views/partials/navbar" .}}
+
+
+
+
+
+
+
+
+ Job Details
+
+
+
Live job status, reasoning traces, and execution details
+
+
+ Back to Jobs
+
+
+
+
+
+
+
+
Job Status
+
+
+
+ Cancel
+
+
+
+
+
+
+
+
+
+
Agent Prompt Template
+
The original prompt template from the task definition.
+
+
+
+
+
+
Cron Parameters
+
Parameters configured for cron-triggered executions of this task.
+
+
+
+
+
+
Job Parameters
+
Parameters used for this specific job execution.
+
+
+
+
+
+
Rendered Job Prompt
+
The prompt with parameters substituted, as it was sent to the agent.
+
+
+
+
+
+
+
+
+
+
+
+
Execution Traces
+
+
+
No execution traces available yet. Traces will appear here as the job executes.
+
+
+
+
+
+
+
Webhook Status
+
+
+
+
+
+
+
+
+
+
+
+
+
Webhook Delivery Errors:
+
+
+
+
+
+
+
+
+
+
+
+
+
+
diff --git a/core/http/views/agent-jobs.html b/core/http/views/agent-jobs.html
new file mode 100644
index 000000000..75724a732
--- /dev/null
+++ b/core/http/views/agent-jobs.html
@@ -0,0 +1,648 @@
+
+
+{{template "views/partials/head" .}}
+
+
+
+
+ {{template "views/partials/navbar" .}}
+
+
+
+
+
+
+
+
+ Agent Jobs
+
+
+
Manage agent tasks and monitor job execution
+
+
+ Create Task
+
+
+
+
+
+
+
+
+
+
+
+
+ No Models Installed
+
+
+
+ To use Agent Jobs, you need to install a model first. Agent Jobs require models with MCP (Model Context Protocol) configuration.
+
+
+
+
+
+
+
+
+
Model Gallery
+
Browse and install pre-configured models
+
+
+
+
+
+
Import Models
+
Upload your own model files
+
+
+
+
+
+
API Download
+
Use the API to download models programmatically
+
+
+
+
+
+
+
+ How to Get Started
+
+
+
+
+ 1
+
+
+
Browse the Model Gallery
+
Explore our curated collection of pre-configured models. Find models for chat, image generation, audio processing, and more.
+
+
+
+
+ 2
+
+
+
Install a Model
+
Click on a model from the gallery to install it, or use the import feature to upload your own model files.
+
+
+
+
+ 3
+
+
+
Configure MCP
+
After installing a model, configure MCP (Model Context Protocol) to enable Agent Jobs functionality.
+
+
+
+
+
+
+
+
+
+
+
+
+
+
+
+
+
+ MCP Configuration Required
+
+
+
+ You have models installed, but none have MCP (Model Context Protocol) enabled. Agent Jobs require MCP to function.
+
+
+
+
+
+
+ Available Models
+
+
+
+
+
+
+
+
+ How to Enable MCP
+
+
+
+
+ 1
+
+
+
Edit a Model Configuration
+
Click "Configure MCP" on any model above, or navigate to the model editor to add MCP configuration.
+
+
+
+
+ 2
+
+
+
Add MCP Configuration
+
In the model YAML, add MCP server or stdio configuration. See the documentation for detailed examples.
+
+
+
+
+ 3
+
+
+
Save and Return
+
After saving the MCP configuration, return to this page to create your first Agent Job task.
+
+
+
+
+
+
+
+
+
+
+
+
Tasks
+
+
+
+
+ Name
+ Model
+ Cron
+ Status
+ Actions
+
+
+
+
+
+
+
+
+
+
+
+
+
+
+ -
+
+
+
+
+
+
+
+
+
+
+
+ No tasks found. Create one
+
+
+
+
+
+
+
+
+
+
+
Job History
+
+
+ All Status
+ Pending
+ Running
+ Completed
+ Failed
+ Cancelled
+
+
+ Clear History
+
+
+
+
+
+
+
+ Job ID
+ Task
+ Status
+ Created
+ Actions
+
+
+
+
+
+
+
+
+
+
+
+
+
+
+
+
+
+
+
+
+
+
+
+ No jobs found
+
+
+
+
+
+
+
+
+
+
+
+
Execute Task
+
+
+
+
+
+
+
+
+
Parameters
+
+ Enter parameters as key-value pairs (one per line, format: key=value).
+ These will be used to template the prompt.
+
+
+
+ Example: user_name=Alice
+
+
+
+
+ Cancel
+
+
+ Execute
+
+
+
+
+
+
+
+
+
+
+
+
+
+
+
diff --git a/core/http/views/agent-task-details.html b/core/http/views/agent-task-details.html
new file mode 100644
index 000000000..e75d8bf24
--- /dev/null
+++ b/core/http/views/agent-task-details.html
@@ -0,0 +1,913 @@
+
+
+{{template "views/partials/head" .}}
+
+
+
+
+ {{template "views/partials/navbar" .}}
+
+
+
+
+
+
+
+
+
+
+ Execute
+
+
+ Edit
+
+
+ Delete
+
+
+
+
+
+
+ Cancel
+
+
+ Save
+
+
+
+
+ Back
+
+
+
+
+
+
+
+
+
+
+
+
+
+
+
+
+
+
API Usage Examples
+
+ Use these curl commands to interact with this task programmatically.
+
+
+
+
+
+
+
+ Execute Task by ID
+
+
+
curl -X POST {{ .BaseURL }}api/agent/jobs/execute \
+ -H "Content-Type: application/json" \
+ -H "Authorization: Bearer YOUR_API_KEY" \
+ -d '{
+ "task_id": " ",
+ "parameters": {
+ "user_name": "Alice",
+ "job_title": "Software Engineer",
+ "task_description": "Review code changes"
+ }
+ }'
+
+
+
+
+
+
+
+ Execute Task by Name
+
+
+
curl -X POST {{ .BaseURL }}api/agent/tasks/ /execute \
+ -H "Content-Type: application/json" \
+ -H "Authorization: Bearer YOUR_API_KEY" \
+ -d '{
+ "user_name": "Bob",
+ "job_title": "Data Scientist",
+ "task_description": "Analyze sales data"
+ }'
+
+
+
+ The request body should be a JSON object where keys are parameter names and values are strings.
+ If no body is provided, the task will execute with empty parameters.
+
+
+
+
+
+
+
+ Check Job Status
+
+
+
curl -X GET {{ .BaseURL }}api/agent/jobs/JOB_ID \
+ -H "Authorization: Bearer YOUR_API_KEY"
+
+
+ After executing a task, you will receive a job_id in the response. Use it to query the job's status and results.
+
+
+
+
+
+
+
+
Webhook Configuration
+
+
+
+
+
+
+
+
+
Job History
+
+
+ All Status
+ Pending
+ Running
+ Completed
+ Failed
+ Cancelled
+
+
+ Clear History
+
+
+
+
+
+
+
+ Job ID
+ Status
+ Created
+ Triggered By
+ Actions
+
+
+
+
+
+
+
+
+
+
+
+
+
+
+
+
+
+
+
+
+
+ No jobs found for this task
+
+
+
+
+
+
+
+
+
+
+
+
+
Execute Task
+
+
+
+
+
+
+
+
+
Parameters
+
+ Enter parameters as key-value pairs (one per line, format: key=value).
+ These will be used to template the prompt.
+
+
+
+ Example: user_name=Alice
+
+
+
+
+ Cancel
+
+
+ Execute
+
+
+
+
+
+
+
+
+
+
+
diff --git a/core/http/views/chat.html b/core/http/views/chat.html
index df2e776e2..a956a9413 100644
--- a/core/http/views/chat.html
+++ b/core/http/views/chat.html
@@ -45,19 +45,29 @@ SOFTWARE.
function __initChatStore() {
if (!window.Alpine) return;
- // Check for MCP mode from localStorage (set by index page)
+ // Check for MCP mode from localStorage (set by index page) or URL parameter
// Note: We don't clear localStorage here - chat.js will handle that after reading all data
let initialMcpMode = false;
- try {
- const chatData = localStorage.getItem('localai_index_chat_data');
- if (chatData) {
- const parsed = JSON.parse(chatData);
- if (parsed.mcpMode === true) {
- initialMcpMode = true;
+
+ // First check URL parameter
+ const urlParams = new URLSearchParams(window.location.search);
+ if (urlParams.get('mcp') === 'true') {
+ initialMcpMode = true;
+ }
+
+ // Then check localStorage (URL param takes precedence)
+ if (!initialMcpMode) {
+ try {
+ const chatData = localStorage.getItem('localai_index_chat_data');
+ if (chatData) {
+ const parsed = JSON.parse(chatData);
+ if (parsed.mcpMode === true) {
+ initialMcpMode = true;
+ }
}
+ } catch (e) {
+ console.error('Error reading MCP mode from localStorage:', e);
}
- } catch (e) {
- console.error('Error reading MCP mode from localStorage:', e);
}
if (Alpine.store("chat")) {
@@ -565,6 +575,13 @@ SOFTWARE.
{{ end }}
{{ end }}
+ {{ if $model }}
+
+
+
+ {{ end }}
API
+
+ Agent Jobs
+
diff --git a/core/http/views/settings.html b/core/http/views/settings.html
index b6acaea0d..88f4baffb 100644
--- a/core/http/views/settings.html
+++ b/core/http/views/settings.html
@@ -317,6 +317,29 @@
+
+
+
+
+ Agent Jobs Settings
+
+
+ Configure agent job retention and cleanup
+
+
+
+
+
+
Job Retention Days
+
Number of days to keep job history (default: 30)
+
+
+
+
+
@@ -455,7 +478,8 @@ function settingsDashboard() {
autoload_backend_galleries: false,
galleries_json: '[]',
backend_galleries_json: '[]',
- api_keys_text: ''
+ api_keys_text: '',
+ agent_job_retention_days: 30
},
sourceInfo: '',
saving: false,
@@ -492,7 +516,8 @@ function settingsDashboard() {
autoload_backend_galleries: data.autoload_backend_galleries || false,
galleries_json: JSON.stringify(data.galleries || [], null, 2),
backend_galleries_json: JSON.stringify(data.backend_galleries || [], null, 2),
- api_keys_text: (data.api_keys || []).join('\n')
+ api_keys_text: (data.api_keys || []).join('\n'),
+ agent_job_retention_days: data.agent_job_retention_days || 30
};
this.sourceInfo = data.source || 'default';
} else {
@@ -609,6 +634,9 @@ function settingsDashboard() {
return;
}
}
+ if (this.settings.agent_job_retention_days !== undefined) {
+ payload.agent_job_retention_days = parseInt(this.settings.agent_job_retention_days) || 30;
+ }
const response = await fetch('/api/settings', {
method: 'POST',
diff --git a/core/schema/agent_jobs.go b/core/schema/agent_jobs.go
new file mode 100644
index 000000000..4d867416a
--- /dev/null
+++ b/core/schema/agent_jobs.go
@@ -0,0 +1,111 @@
+package schema
+
+import (
+ "time"
+)
+
+// Task represents a reusable agent task definition
+type Task struct {
+ ID string `json:"id"` // UUID
+ Name string `json:"name"` // User-friendly name
+ Description string `json:"description"` // Optional description
+ Model string `json:"model"` // Model name (must have MCP config)
+ Prompt string `json:"prompt"` // Template prompt (supports {{.param}} syntax)
+ CreatedAt time.Time `json:"created_at"`
+ UpdatedAt time.Time `json:"updated_at"`
+ Enabled bool `json:"enabled"` // Can be disabled without deletion
+ Cron string `json:"cron,omitempty"` // Optional cron expression
+ CronParameters map[string]string `json:"cron_parameters,omitempty"` // Parameters to use when executing cron jobs
+
+ // Webhook configuration (for notifications)
+ // Support multiple webhook endpoints
+ // Webhooks can handle both success and failure cases using template variables:
+ // - {{.Job}} - Job object with all fields
+ // - {{.Task}} - Task object
+ // - {{.Result}} - Job result (if successful)
+ // - {{.Error}} - Error message (if failed, empty string if successful)
+ // - {{.Status}} - Job status string
+ Webhooks []WebhookConfig `json:"webhooks,omitempty"` // Webhook configs for job completion notifications
+
+}
+
+// WebhookConfig represents configuration for sending webhook notifications
+type WebhookConfig struct {
+ URL string `json:"url"` // Webhook endpoint URL
+ Method string `json:"method"` // HTTP method (POST, PUT, PATCH) - default: POST
+ Headers map[string]string `json:"headers,omitempty"` // Custom headers (e.g., Authorization)
+ PayloadTemplate string `json:"payload_template,omitempty"` // Optional template for payload
+ // If PayloadTemplate is empty, uses default JSON structure
+ // Available template variables:
+ // - {{.Job}} - Job object with all fields
+ // - {{.Task}} - Task object
+ // - {{.Result}} - Job result (if successful)
+ // - {{.Error}} - Error message (if failed, empty string if successful)
+ // - {{.Status}} - Job status string
+}
+
+// JobStatus represents the status of a job
+type JobStatus string
+
+const (
+ JobStatusPending JobStatus = "pending"
+ JobStatusRunning JobStatus = "running"
+ JobStatusCompleted JobStatus = "completed"
+ JobStatusFailed JobStatus = "failed"
+ JobStatusCancelled JobStatus = "cancelled"
+)
+
+// Job represents a single execution instance of a task
+type Job struct {
+ ID string `json:"id"` // UUID
+ TaskID string `json:"task_id"` // Reference to Task
+ Status JobStatus `json:"status"` // pending, running, completed, failed, cancelled
+ Parameters map[string]string `json:"parameters"` // Template parameters
+ Result string `json:"result,omitempty"` // Agent response
+ Error string `json:"error,omitempty"` // Error message if failed
+ StartedAt *time.Time `json:"started_at,omitempty"`
+ CompletedAt *time.Time `json:"completed_at,omitempty"`
+ CreatedAt time.Time `json:"created_at"`
+ TriggeredBy string `json:"triggered_by"` // "manual", "cron", "api"
+
+ // Webhook delivery tracking
+ WebhookSent bool `json:"webhook_sent,omitempty"`
+ WebhookSentAt *time.Time `json:"webhook_sent_at,omitempty"`
+ WebhookError string `json:"webhook_error,omitempty"` // Error if webhook failed
+
+ // Execution traces (reasoning, tool calls, tool results)
+ Traces []JobTrace `json:"traces,omitempty"`
+}
+
+// JobTrace represents a single execution trace entry
+type JobTrace struct {
+ Type string `json:"type"` // "reasoning", "tool_call", "tool_result", "status"
+ Content string `json:"content"` // The actual trace content
+ Timestamp time.Time `json:"timestamp"` // When this trace occurred
+ ToolName string `json:"tool_name,omitempty"` // Tool name (for tool_call/tool_result)
+ Arguments map[string]interface{} `json:"arguments,omitempty"` // Tool arguments or result data
+}
+
+// JobExecutionRequest represents a request to execute a job
+type JobExecutionRequest struct {
+ TaskID string `json:"task_id"` // Required
+ Parameters map[string]string `json:"parameters"` // Optional, for templating
+}
+
+// JobExecutionResponse represents the response after creating a job
+type JobExecutionResponse struct {
+ JobID string `json:"job_id"`
+ Status string `json:"status"`
+ URL string `json:"url"` // URL to check job status
+}
+
+// TasksFile represents the structure of agent_tasks.json
+type TasksFile struct {
+ Tasks []Task `json:"tasks"`
+}
+
+// JobsFile represents the structure of agent_jobs.json
+type JobsFile struct {
+ Jobs []Job `json:"jobs"`
+ LastCleanup time.Time `json:"last_cleanup,omitempty"`
+}
diff --git a/core/services/agent_jobs.go b/core/services/agent_jobs.go
new file mode 100644
index 000000000..1702e6d6d
--- /dev/null
+++ b/core/services/agent_jobs.go
@@ -0,0 +1,1180 @@
+package services
+
+import (
+ "bytes"
+ "context"
+ "encoding/json"
+ "errors"
+ "fmt"
+ "io"
+ "net"
+ "net/http"
+ "os"
+ "path/filepath"
+ "sort"
+ "strings"
+ "sync"
+ "text/template"
+ "time"
+
+ "github.com/Masterminds/sprig/v3"
+ "github.com/google/uuid"
+ "github.com/mudler/LocalAI/core/config"
+ mcpTools "github.com/mudler/LocalAI/core/http/endpoints/mcp"
+ "github.com/mudler/LocalAI/core/schema"
+ "github.com/mudler/LocalAI/core/templates"
+ "github.com/mudler/LocalAI/pkg/model"
+ "github.com/mudler/LocalAI/pkg/xsync"
+ "github.com/mudler/cogito"
+ "github.com/robfig/cron/v3"
+ "github.com/rs/zerolog/log"
+)
+
+// AgentJobService manages agent tasks and job execution
+type AgentJobService struct {
+ appConfig *config.ApplicationConfig
+ modelLoader *model.ModelLoader
+ configLoader *config.ModelConfigLoader
+ evaluator *templates.Evaluator
+
+ // Storage (file-based with in-memory cache)
+ tasks *xsync.SyncedMap[string, schema.Task]
+ jobs *xsync.SyncedMap[string, schema.Job]
+ tasksFile string // Path to agent_tasks.json
+ jobsFile string // Path to agent_jobs.json
+
+ // Job execution channel
+ jobQueue chan JobExecution
+
+ // Cancellation support
+ cancellations *xsync.SyncedMap[string, context.CancelFunc]
+
+ // Cron scheduler
+ cronScheduler *cron.Cron
+ cronEntries *xsync.SyncedMap[string, cron.EntryID]
+
+ // Job retention
+ retentionDays int // From runtime settings, default: 30
+
+ // Service lifecycle
+ ctx context.Context
+ cancel context.CancelFunc
+
+ // Mutex for file operations
+ fileMutex sync.Mutex
+}
+
+// JobExecution represents a job to be executed
+type JobExecution struct {
+ Job schema.Job
+ Task schema.Task
+ Ctx context.Context
+ Cancel context.CancelFunc
+}
+
+// NewAgentJobService creates a new AgentJobService instance
+func NewAgentJobService(
+ appConfig *config.ApplicationConfig,
+ modelLoader *model.ModelLoader,
+ configLoader *config.ModelConfigLoader,
+ evaluator *templates.Evaluator,
+) *AgentJobService {
+ retentionDays := appConfig.AgentJobRetentionDays
+ if retentionDays == 0 {
+ retentionDays = 30 // Default
+ }
+
+ tasksFile := ""
+ jobsFile := ""
+ if appConfig.DynamicConfigsDir != "" {
+ tasksFile = filepath.Join(appConfig.DynamicConfigsDir, "agent_tasks.json")
+ jobsFile = filepath.Join(appConfig.DynamicConfigsDir, "agent_jobs.json")
+ }
+
+ return &AgentJobService{
+ appConfig: appConfig,
+ modelLoader: modelLoader,
+ configLoader: configLoader,
+ evaluator: evaluator,
+ tasks: xsync.NewSyncedMap[string, schema.Task](),
+ jobs: xsync.NewSyncedMap[string, schema.Job](),
+ tasksFile: tasksFile,
+ jobsFile: jobsFile,
+ jobQueue: make(chan JobExecution, 100), // Buffer for 100 jobs
+ cancellations: xsync.NewSyncedMap[string, context.CancelFunc](),
+ cronScheduler: cron.New(), // Support seconds in cron
+ cronEntries: xsync.NewSyncedMap[string, cron.EntryID](),
+ retentionDays: retentionDays,
+ }
+}
+
+// LoadTasksFromFile loads tasks from agent_tasks.json
+func (s *AgentJobService) LoadTasksFromFile() error {
+ if s.tasksFile == "" {
+ return nil // No file path configured
+ }
+
+ s.fileMutex.Lock()
+ defer s.fileMutex.Unlock()
+
+ if _, err := os.Stat(s.tasksFile); os.IsNotExist(err) {
+ log.Debug().Msg("agent_tasks.json not found, starting with empty tasks")
+ return nil
+ }
+
+ fileContent, err := os.ReadFile(s.tasksFile)
+ if err != nil {
+ return fmt.Errorf("failed to read tasks file: %w", err)
+ }
+
+ var tasksFile schema.TasksFile
+ if err := json.Unmarshal(fileContent, &tasksFile); err != nil {
+ return fmt.Errorf("failed to parse tasks file: %w", err)
+ }
+
+ for _, task := range tasksFile.Tasks {
+ s.tasks.Set(task.ID, task)
+ // Schedule cron if enabled and has cron expression
+ if task.Enabled && task.Cron != "" {
+ if err := s.ScheduleCronTask(task); err != nil {
+ log.Warn().Err(err).Str("task_id", task.ID).Msg("Failed to schedule cron task on load")
+ }
+ }
+ }
+
+ log.Info().Int("count", len(tasksFile.Tasks)).Msg("Loaded tasks from file")
+
+ return nil
+}
+
+// SaveTasksToFile saves tasks to agent_tasks.json
+func (s *AgentJobService) SaveTasksToFile() error {
+ if s.tasksFile == "" {
+ return nil // No file path configured
+ }
+
+ s.fileMutex.Lock()
+ defer s.fileMutex.Unlock()
+
+ tasksFile := schema.TasksFile{
+ Tasks: s.tasks.Values(),
+ }
+
+ fileContent, err := json.MarshalIndent(tasksFile, "", " ")
+ if err != nil {
+ return fmt.Errorf("failed to marshal tasks: %w", err)
+ }
+
+ if err := os.WriteFile(s.tasksFile, fileContent, 0600); err != nil {
+ return fmt.Errorf("failed to write tasks file: %w", err)
+ }
+
+ return nil
+}
+
+// LoadJobsFromFile loads jobs from agent_jobs.json
+func (s *AgentJobService) LoadJobsFromFile() error {
+ if s.jobsFile == "" {
+ return nil // No file path configured
+ }
+
+ s.fileMutex.Lock()
+ defer s.fileMutex.Unlock()
+
+ if _, err := os.Stat(s.jobsFile); os.IsNotExist(err) {
+ log.Debug().Msg("agent_jobs.json not found, starting with empty jobs")
+ return nil
+ }
+
+ fileContent, err := os.ReadFile(s.jobsFile)
+ if err != nil {
+ return fmt.Errorf("failed to read jobs file: %w", err)
+ }
+
+ var jobsFile schema.JobsFile
+ if err := json.Unmarshal(fileContent, &jobsFile); err != nil {
+ return fmt.Errorf("failed to parse jobs file: %w", err)
+ }
+
+ // Load jobs into memory
+ for _, job := range jobsFile.Jobs {
+ s.jobs.Set(job.ID, job)
+ }
+
+ log.Info().Int("count", len(jobsFile.Jobs)).Msg("Loaded jobs from file")
+ return nil
+}
+
+// SaveJobsToFile saves jobs to agent_jobs.json
+func (s *AgentJobService) SaveJobsToFile() error {
+ if s.jobsFile == "" {
+ return nil // No file path configured
+ }
+
+ s.fileMutex.Lock()
+ defer s.fileMutex.Unlock()
+
+ jobsFile := schema.JobsFile{
+ Jobs: s.jobs.Values(),
+ LastCleanup: time.Now(),
+ }
+
+ fileContent, err := json.MarshalIndent(jobsFile, "", " ")
+ if err != nil {
+ return fmt.Errorf("failed to marshal jobs: %w", err)
+ }
+
+ if err := os.WriteFile(s.jobsFile, fileContent, 0600); err != nil {
+ return fmt.Errorf("failed to write jobs file: %w", err)
+ }
+
+ return nil
+}
+
+// CreateTask creates a new task
+func (s *AgentJobService) CreateTask(task schema.Task) (string, error) {
+ if task.Name == "" {
+ return "", fmt.Errorf("task name is required")
+ }
+ if task.Model == "" {
+ return "", fmt.Errorf("task model is required")
+ }
+ if task.Prompt == "" {
+ return "", fmt.Errorf("task prompt is required")
+ }
+
+ // Generate UUID
+ id := uuid.New().String()
+ task.ID = id
+ now := time.Now()
+ task.CreatedAt = now
+ task.UpdatedAt = now
+ if !task.Enabled {
+ task.Enabled = true // Default to enabled
+ }
+
+ // Store task
+ s.tasks.Set(id, task)
+
+ // Schedule cron if enabled and has cron expression
+ if task.Enabled && task.Cron != "" {
+ if err := s.ScheduleCronTask(task); err != nil {
+ log.Warn().Err(err).Str("task_id", id).Msg("Failed to schedule cron task")
+ // Don't fail task creation if cron scheduling fails
+ }
+ }
+
+ // Save to file
+ if err := s.SaveTasksToFile(); err != nil {
+ log.Error().Err(err).Msg("Failed to save tasks to file")
+ // Don't fail task creation if file save fails
+ }
+
+ return id, nil
+}
+
+// UpdateTask updates an existing task
+func (s *AgentJobService) UpdateTask(id string, task schema.Task) error {
+ if !s.tasks.Exists(id) {
+ return fmt.Errorf("task not found: %s", id)
+ }
+ existing := s.tasks.Get(id)
+
+ // Preserve ID and CreatedAt
+ task.ID = id
+ task.CreatedAt = existing.CreatedAt
+ task.UpdatedAt = time.Now()
+
+ // Unschedule old cron if it had one
+ if existing.Cron != "" {
+ s.UnscheduleCronTask(id)
+ }
+
+ // Store updated task
+ s.tasks.Set(id, task)
+
+ // Schedule new cron if enabled and has cron expression
+ if task.Enabled && task.Cron != "" {
+ if err := s.ScheduleCronTask(task); err != nil {
+ log.Warn().Err(err).Str("task_id", id).Msg("Failed to schedule cron task")
+ }
+ }
+
+ // Save to file
+ if err := s.SaveTasksToFile(); err != nil {
+ log.Error().Err(err).Msg("Failed to save tasks to file")
+ }
+
+ return nil
+}
+
+// DeleteTask deletes a task
+func (s *AgentJobService) DeleteTask(id string) error {
+ if !s.tasks.Exists(id) {
+ return fmt.Errorf("task not found: %s", id)
+ }
+
+ // Unschedule cron
+ s.UnscheduleCronTask(id)
+
+ // Remove from memory
+ s.tasks.Delete(id)
+
+ // Save to file
+ if err := s.SaveTasksToFile(); err != nil {
+ log.Error().Err(err).Msg("Failed to save tasks to file")
+ }
+
+ return nil
+}
+
+// GetTask retrieves a task by ID
+func (s *AgentJobService) GetTask(id string) (*schema.Task, error) {
+ task := s.tasks.Get(id)
+ if task.ID == "" {
+ return nil, fmt.Errorf("task not found: %s", id)
+ }
+ return &task, nil
+}
+
+// ListTasks returns all tasks, sorted by creation date (newest first)
+func (s *AgentJobService) ListTasks() []schema.Task {
+ tasks := s.tasks.Values()
+ // Sort by CreatedAt descending (newest first), then by Name for stability
+ sort.Slice(tasks, func(i, j int) bool {
+ if tasks[i].CreatedAt.Equal(tasks[j].CreatedAt) {
+ return tasks[i].Name < tasks[j].Name
+ }
+ return tasks[i].CreatedAt.After(tasks[j].CreatedAt)
+ })
+ return tasks
+}
+
+// buildPrompt builds a prompt from a template with parameters
+func (s *AgentJobService) buildPrompt(templateStr string, params map[string]string) (string, error) {
+ tmpl, err := template.New("prompt").Parse(templateStr)
+ if err != nil {
+ return "", fmt.Errorf("failed to parse prompt template: %w", err)
+ }
+
+ var buf bytes.Buffer
+ if err := tmpl.Execute(&buf, params); err != nil {
+ return "", fmt.Errorf("failed to execute prompt template: %w", err)
+ }
+
+ return buf.String(), nil
+}
+
+// ExecuteJob creates and queues a job for execution
+func (s *AgentJobService) ExecuteJob(taskID string, params map[string]string, triggeredBy string) (string, error) {
+ task := s.tasks.Get(taskID)
+ if task.ID == "" {
+ return "", fmt.Errorf("task not found: %s", taskID)
+ }
+
+ if !task.Enabled {
+ return "", fmt.Errorf("task is disabled: %s", taskID)
+ }
+
+ // Create job
+ jobID := uuid.New().String()
+ now := time.Now()
+ job := schema.Job{
+ ID: jobID,
+ TaskID: taskID,
+ Status: schema.JobStatusPending,
+ Parameters: params,
+ CreatedAt: now,
+ TriggeredBy: triggeredBy,
+ }
+
+ // Store job
+ s.jobs.Set(jobID, job)
+
+ // Save to file (async, don't block)
+ go func() {
+ if err := s.SaveJobsToFile(); err != nil {
+ log.Error().Err(err).Msg("Failed to save jobs to file")
+ }
+ }()
+
+ // Create context for cancellation
+ ctx, cancel := context.WithCancel(context.Background())
+ s.cancellations.Set(jobID, cancel)
+
+ // Queue job
+ select {
+ case s.jobQueue <- JobExecution{
+ Job: job,
+ Task: task,
+ Ctx: ctx,
+ Cancel: cancel,
+ }:
+ default:
+ // Queue is full, update job status
+ job.Status = schema.JobStatusFailed
+ job.Error = "job queue is full"
+ s.jobs.Set(jobID, job)
+ return "", fmt.Errorf("job queue is full")
+ }
+
+ return jobID, nil
+}
+
+// GetJob retrieves a job by ID
+func (s *AgentJobService) GetJob(id string) (*schema.Job, error) {
+ job := s.jobs.Get(id)
+ if job.ID == "" {
+ return nil, fmt.Errorf("job not found: %s", id)
+ }
+ return &job, nil
+}
+
+// ListJobs returns jobs, optionally filtered by task_id and status
+func (s *AgentJobService) ListJobs(taskID *string, status *schema.JobStatus, limit int) []schema.Job {
+ allJobs := s.jobs.Values()
+ filtered := []schema.Job{}
+
+ for _, job := range allJobs {
+ if taskID != nil && job.TaskID != *taskID {
+ continue
+ }
+ if status != nil && job.Status != *status {
+ continue
+ }
+ filtered = append(filtered, job)
+ }
+
+ // Sort by CreatedAt descending (newest first)
+ for i := 0; i < len(filtered)-1; i++ {
+ for j := i + 1; j < len(filtered); j++ {
+ if filtered[i].CreatedAt.Before(filtered[j].CreatedAt) {
+ filtered[i], filtered[j] = filtered[j], filtered[i]
+ }
+ }
+ }
+
+ // Apply limit
+ if limit > 0 && limit < len(filtered) {
+ filtered = filtered[:limit]
+ }
+
+ return filtered
+}
+
+// CancelJob cancels a running job
+func (s *AgentJobService) CancelJob(id string) error {
+ job := s.jobs.Get(id)
+ if job.ID == "" {
+ return fmt.Errorf("job not found: %s", id)
+ }
+
+ if job.Status != schema.JobStatusPending && job.Status != schema.JobStatusRunning {
+ return fmt.Errorf("job cannot be cancelled: status is %s", job.Status)
+ }
+
+ // Cancel context
+ if s.cancellations.Exists(id) {
+ cancel := s.cancellations.Get(id)
+ cancel()
+ s.cancellations.Delete(id)
+ }
+
+ // Update job status
+ now := time.Now()
+ job.Status = schema.JobStatusCancelled
+ job.CompletedAt = &now
+ s.jobs.Set(id, job)
+
+ // Save to file (async)
+ go func() {
+ if err := s.SaveJobsToFile(); err != nil {
+ log.Error().Err(err).Msg("Failed to save jobs to file")
+ }
+ }()
+
+ return nil
+}
+
+// DeleteJob deletes a job
+func (s *AgentJobService) DeleteJob(id string) error {
+ if !s.jobs.Exists(id) {
+ return fmt.Errorf("job not found: %s", id)
+ }
+
+ s.jobs.Delete(id)
+
+ // Save to file
+ if err := s.SaveJobsToFile(); err != nil {
+ log.Error().Err(err).Msg("Failed to save jobs to file")
+ }
+
+ return nil
+}
+
+// executeJobInternal executes a job using cogito
+func (s *AgentJobService) executeJobInternal(job schema.Job, task schema.Task, ctx context.Context) error {
+ // Update job status to running
+ now := time.Now()
+ job.Status = schema.JobStatusRunning
+ job.StartedAt = &now
+ s.jobs.Set(job.ID, job)
+
+ // Load model config
+ modelConfig, err := s.configLoader.LoadModelConfigFileByNameDefaultOptions(task.Model, s.appConfig)
+ if err != nil {
+ job.Status = schema.JobStatusFailed
+ job.Error = fmt.Sprintf("failed to load model config: %v", err)
+ completedAt := time.Now()
+ job.CompletedAt = &completedAt
+ s.jobs.Set(job.ID, job)
+ return fmt.Errorf("failed to load model config: %w", err)
+ }
+
+ // Validate MCP configuration
+ if modelConfig.MCP.Servers == "" && modelConfig.MCP.Stdio == "" {
+ job.Status = schema.JobStatusFailed
+ job.Error = "no MCP servers configured for model"
+ completedAt := time.Now()
+ job.CompletedAt = &completedAt
+ s.jobs.Set(job.ID, job)
+ return fmt.Errorf("no MCP servers configured for model: %s", task.Model)
+ }
+
+ // Get MCP config from model config
+ remote, stdio, err := modelConfig.MCP.MCPConfigFromYAML()
+ if err != nil {
+ job.Status = schema.JobStatusFailed
+ job.Error = fmt.Sprintf("failed to get MCP config: %v", err)
+ completedAt := time.Now()
+ job.CompletedAt = &completedAt
+ s.jobs.Set(job.ID, job)
+ return fmt.Errorf("failed to get MCP config: %w", err)
+ }
+
+ // Get MCP sessions
+ sessions, err := mcpTools.SessionsFromMCPConfig(modelConfig.Name, remote, stdio)
+ if err != nil {
+ job.Status = schema.JobStatusFailed
+ job.Error = fmt.Sprintf("failed to get MCP sessions: %v", err)
+ completedAt := time.Now()
+ job.CompletedAt = &completedAt
+ s.jobs.Set(job.ID, job)
+ return fmt.Errorf("failed to get MCP sessions: %w", err)
+ }
+
+ if len(sessions) == 0 {
+ job.Status = schema.JobStatusFailed
+ job.Error = "no working MCP servers found"
+ completedAt := time.Now()
+ job.CompletedAt = &completedAt
+ s.jobs.Set(job.ID, job)
+ return fmt.Errorf("no working MCP servers found")
+ }
+
+ // Build prompt from template
+ prompt, err := s.buildPrompt(task.Prompt, job.Parameters)
+ if err != nil {
+ job.Status = schema.JobStatusFailed
+ job.Error = fmt.Sprintf("failed to build prompt: %v", err)
+ completedAt := time.Now()
+ job.CompletedAt = &completedAt
+ s.jobs.Set(job.ID, job)
+ return fmt.Errorf("failed to build prompt: %w", err)
+ }
+
+ // Create cogito fragment
+ fragment := cogito.NewEmptyFragment()
+ fragment = fragment.AddMessage("user", prompt)
+
+ // Get API address and key
+ _, port, err := net.SplitHostPort(s.appConfig.APIAddress)
+ if err != nil {
+ job.Status = schema.JobStatusFailed
+ job.Error = fmt.Sprintf("failed to parse API address: %v", err)
+ completedAt := time.Now()
+ job.CompletedAt = &completedAt
+ s.jobs.Set(job.ID, job)
+ return fmt.Errorf("failed to parse API address: %w", err)
+ }
+
+ apiKey := ""
+ if len(s.appConfig.ApiKeys) > 0 {
+ apiKey = s.appConfig.ApiKeys[0]
+ }
+
+ // Create LLM client
+ defaultLLM := cogito.NewOpenAILLM(modelConfig.Name, apiKey, "http://127.0.0.1:"+port)
+
+ // Initialize traces slice
+ job.Traces = []schema.JobTrace{}
+
+ // Build cogito options
+ cogitoOpts := modelConfig.BuildCogitoOptions()
+ cogitoOpts = append(
+ cogitoOpts,
+ cogito.WithContext(ctx),
+ cogito.WithMCPs(sessions...),
+ cogito.WithStatusCallback(func(status string) {
+ log.Debug().Str("job_id", job.ID).Str("model", modelConfig.Name).Msgf("Status: %s", status)
+ // Store trace
+ trace := schema.JobTrace{
+ Type: "status",
+ Content: status,
+ Timestamp: time.Now(),
+ }
+ job.Traces = append(job.Traces, trace)
+ s.jobs.Set(job.ID, job)
+ }),
+ cogito.WithReasoningCallback(func(reasoning string) {
+ log.Debug().Str("job_id", job.ID).Str("model", modelConfig.Name).Msgf("Reasoning: %s", reasoning)
+ // Store trace
+ trace := schema.JobTrace{
+ Type: "reasoning",
+ Content: reasoning,
+ Timestamp: time.Now(),
+ }
+ job.Traces = append(job.Traces, trace)
+ s.jobs.Set(job.ID, job)
+ }),
+ cogito.WithToolCallBack(func(t *cogito.ToolChoice) bool {
+ log.Debug().Str("job_id", job.ID).Str("model", modelConfig.Name).
+ Str("tool", t.Name).Str("reasoning", t.Reasoning).Interface("arguments", t.Arguments).
+ Msg("Tool call")
+ // Store trace
+ arguments := make(map[string]interface{})
+ if t.Arguments != nil {
+ arguments = t.Arguments
+ }
+ trace := schema.JobTrace{
+ Type: "tool_call",
+ Content: t.Reasoning,
+ Timestamp: time.Now(),
+ ToolName: t.Name,
+ Arguments: arguments,
+ }
+ job.Traces = append(job.Traces, trace)
+ s.jobs.Set(job.ID, job)
+ return true
+ }),
+ cogito.WithToolCallResultCallback(func(t cogito.ToolStatus) {
+ log.Debug().Str("job_id", job.ID).Str("model", modelConfig.Name).
+ Str("tool", t.Name).Str("result", t.Result).Interface("tool_arguments", t.ToolArguments).
+ Msg("Tool call result")
+ // Store trace
+ arguments := make(map[string]interface{})
+ // Convert ToolArguments to map via JSON marshaling
+ if toolArgsBytes, err := json.Marshal(t.ToolArguments); err == nil {
+ var toolArgsMap map[string]interface{}
+ if err := json.Unmarshal(toolArgsBytes, &toolArgsMap); err == nil {
+ arguments = toolArgsMap
+ }
+ }
+ arguments["result"] = t.Result
+ trace := schema.JobTrace{
+ Type: "tool_result",
+ Content: t.Result,
+ Timestamp: time.Now(),
+ ToolName: t.Name,
+ Arguments: arguments,
+ }
+ job.Traces = append(job.Traces, trace)
+ s.jobs.Set(job.ID, job)
+ }),
+ )
+
+ // Execute tools
+ f, err := cogito.ExecuteTools(defaultLLM, fragment, cogitoOpts...)
+ if err != nil && !errors.Is(err, cogito.ErrNoToolSelected) {
+ job.Status = schema.JobStatusFailed
+ job.Error = fmt.Sprintf("failed to execute tools: %v", err)
+ completedAt := time.Now()
+ job.CompletedAt = &completedAt
+ s.jobs.Set(job.ID, job)
+ return fmt.Errorf("failed to execute tools: %w", err)
+ }
+
+ // Get final response
+ f, err = defaultLLM.Ask(ctx, f)
+ if err != nil {
+ job.Status = schema.JobStatusFailed
+ job.Error = fmt.Sprintf("failed to get response: %v", err)
+ completedAt := time.Now()
+ job.CompletedAt = &completedAt
+ s.jobs.Set(job.ID, job)
+ return fmt.Errorf("failed to get response: %w", err)
+ }
+
+ // Extract traces from fragment.Status after execution
+ // This provides complete information about tool calls and results
+ // We use Status data to supplement/replace callback data for completeness
+ if f.Status != nil {
+ // Clear existing tool_call and tool_result traces (from callbacks) and replace with Status data
+ // Keep status and reasoning traces from callbacks
+ filteredTraces := []schema.JobTrace{}
+ for _, trace := range job.Traces {
+ if trace.Type != "tool_call" && trace.Type != "tool_result" {
+ filteredTraces = append(filteredTraces, trace)
+ }
+ }
+ job.Traces = filteredTraces
+
+ // Extract tool calls from Status.ToolsCalled
+ if len(f.Status.ToolsCalled) > 0 {
+ for _, toolCallInterface := range f.Status.ToolsCalled {
+ // Marshal to JSON and unmarshal to extract fields
+ if toolCallBytes, err := json.Marshal(toolCallInterface); err == nil {
+ var toolCallData map[string]interface{}
+ if err := json.Unmarshal(toolCallBytes, &toolCallData); err == nil {
+ arguments := make(map[string]interface{})
+ if args, ok := toolCallData["arguments"].(map[string]interface{}); ok {
+ arguments = args
+ }
+ reasoning := ""
+ if r, ok := toolCallData["reasoning"].(string); ok {
+ reasoning = r
+ }
+ name := ""
+ if n, ok := toolCallData["name"].(string); ok {
+ name = n
+ }
+ trace := schema.JobTrace{
+ Type: "tool_call",
+ Content: reasoning,
+ Timestamp: time.Now(),
+ ToolName: name,
+ Arguments: arguments,
+ }
+ job.Traces = append(job.Traces, trace)
+ }
+ }
+ }
+ }
+
+ // Extract tool results from Status.ToolResults
+ if len(f.Status.ToolResults) > 0 {
+ for _, toolResult := range f.Status.ToolResults {
+ arguments := make(map[string]interface{})
+ // Convert ToolArguments to map via JSON marshaling
+ if toolArgsBytes, err := json.Marshal(toolResult.ToolArguments); err == nil {
+ var toolArgsMap map[string]interface{}
+ if err := json.Unmarshal(toolArgsBytes, &toolArgsMap); err == nil {
+ arguments = toolArgsMap
+ }
+ }
+ arguments["result"] = toolResult.Result
+ trace := schema.JobTrace{
+ Type: "tool_result",
+ Content: toolResult.Result,
+ Timestamp: time.Now(),
+ ToolName: toolResult.Name,
+ Arguments: arguments,
+ }
+ job.Traces = append(job.Traces, trace)
+ }
+ }
+ }
+
+ // Update job with result
+ completedAt := time.Now()
+ job.Status = schema.JobStatusCompleted
+ job.Result = f.LastMessage().Content
+ job.CompletedAt = &completedAt
+ s.jobs.Set(job.ID, job)
+
+ // Save to file (async)
+ go func() {
+ if err := s.SaveJobsToFile(); err != nil {
+ log.Error().Err(err).Msg("Failed to save jobs to file")
+ }
+ }()
+
+ // Send webhooks (non-blocking)
+ go func() {
+ s.sendWebhooks(job, task)
+ }()
+
+ return nil
+}
+
+// worker processes jobs from the queue
+func (s *AgentJobService) worker(ctx context.Context) {
+ for {
+ select {
+ case <-ctx.Done():
+ return
+ case exec := <-s.jobQueue:
+ // Check if job was cancelled before execution
+ select {
+ case <-exec.Ctx.Done():
+ job := exec.Job
+ now := time.Now()
+ job.Status = schema.JobStatusCancelled
+ job.CompletedAt = &now
+ s.jobs.Set(job.ID, job)
+ s.cancellations.Delete(job.ID)
+ continue
+ default:
+ }
+
+ // Execute job
+ err := s.executeJobInternal(exec.Job, exec.Task, exec.Ctx)
+ if err != nil {
+ log.Error().Err(err).Str("job_id", exec.Job.ID).Msg("Job execution failed")
+ }
+
+ // Clean up cancellation
+ s.cancellations.Delete(exec.Job.ID)
+ }
+ }
+}
+
+// ScheduleCronTask schedules a task to run on a cron schedule
+func (s *AgentJobService) ScheduleCronTask(task schema.Task) error {
+ if task.Cron == "" {
+ return nil // No cron expression
+ }
+
+ // Parse cron expression (support standard 5-field format)
+ // Convert to 6-field format if needed (with seconds)
+ cronExpr := task.Cron
+ // Use cron parameters if provided, otherwise use empty map
+ cronParams := task.CronParameters
+ if cronParams == nil {
+ cronParams = map[string]string{}
+ }
+ entryID, err := s.cronScheduler.AddFunc(cronExpr, func() {
+ // Create job for cron execution with configured parameters
+ _, err := s.ExecuteJob(task.ID, cronParams, "cron")
+ if err != nil {
+ log.Error().Err(err).Str("task_id", task.ID).Msg("Failed to execute cron job")
+ }
+ })
+ if err != nil {
+ return fmt.Errorf("failed to parse cron expression: %w", err)
+ }
+
+ s.cronEntries.Set(task.ID, entryID)
+ log.Info().Str("task_id", task.ID).Str("cron", cronExpr).Msg("Scheduled cron task")
+ return nil
+}
+
+// UnscheduleCronTask removes a task from the cron scheduler
+func (s *AgentJobService) UnscheduleCronTask(taskID string) {
+ if s.cronEntries.Exists(taskID) {
+ entryID := s.cronEntries.Get(taskID)
+ s.cronScheduler.Remove(entryID)
+ s.cronEntries.Delete(taskID)
+ log.Info().Str("task_id", taskID).Msg("Unscheduled cron task")
+ }
+}
+
+// sendWebhooks sends webhook notifications to all configured webhooks
+func (s *AgentJobService) sendWebhooks(job schema.Job, task schema.Task) {
+ // Collect all webhook configs from new format
+ webhookConfigs := task.Webhooks
+
+ if len(webhookConfigs) == 0 {
+ return // No webhooks configured
+ }
+
+ log.Info().Str("job_id", job.ID).Int("webhook_count", len(webhookConfigs)).Msg("Sending webhooks")
+
+ // Send all webhooks concurrently and track results
+ var wg sync.WaitGroup
+ errors := make(chan webhookError, len(webhookConfigs))
+ successCount := 0
+
+ for _, webhookConfig := range webhookConfigs {
+ wg.Add(1)
+ go func(config schema.WebhookConfig) {
+ defer wg.Done()
+ if err := s.sendWebhook(job, task, config); err != nil {
+ errors <- webhookError{
+ URL: config.URL,
+ Error: err.Error(),
+ }
+ } else {
+ successCount++
+ }
+ }(webhookConfig)
+ }
+ wg.Wait()
+ close(errors)
+
+ // Collect errors
+ var webhookErrors []string
+ for err := range errors {
+ webhookErrors = append(webhookErrors, fmt.Sprintf("%s: %s", err.URL, err.Error))
+ }
+
+ // Update job with webhook status
+ job = s.jobs.Get(job.ID)
+ if job.ID == "" {
+ return
+ }
+
+ now := time.Now()
+ if len(webhookErrors) == 0 {
+ // All webhooks succeeded
+ job.WebhookSent = true
+ job.WebhookSentAt = &now
+ job.WebhookError = ""
+ } else if successCount > 0 {
+ // Some succeeded, some failed
+ job.WebhookSent = true
+ job.WebhookSentAt = &now
+ job.WebhookError = fmt.Sprintf("Some webhooks failed (%d/%d succeeded): %s", successCount, len(webhookConfigs), strings.Join(webhookErrors, "; "))
+ } else {
+ // All failed
+ job.WebhookSent = false
+ job.WebhookError = fmt.Sprintf("All webhooks failed: %s", strings.Join(webhookErrors, "; "))
+ }
+
+ s.jobs.Set(job.ID, job)
+
+ // Save to file (async)
+ go func() {
+ if err := s.SaveJobsToFile(); err != nil {
+ log.Error().Err(err).Msg("Failed to save jobs to file")
+ }
+ }()
+}
+
+// webhookError represents a webhook delivery error
+type webhookError struct {
+ URL string
+ Error string
+}
+
+// sendWebhook sends a single webhook notification
+// Returns an error if the webhook delivery failed
+func (s *AgentJobService) sendWebhook(job schema.Job, task schema.Task, webhookConfig schema.WebhookConfig) error {
+ // Build payload
+ payload, err := s.buildWebhookPayload(job, task, webhookConfig)
+ if err != nil {
+ log.Error().Err(err).Str("job_id", job.ID).Str("webhook_url", webhookConfig.URL).Msg("Failed to build webhook payload")
+ return fmt.Errorf("failed to build payload: %w", err)
+ }
+
+ log.Debug().Str("job_id", job.ID).Str("webhook_url", webhookConfig.URL).Str("payload", string(payload)).Msg("Sending webhook")
+
+ // Determine HTTP method (default to POST)
+ method := webhookConfig.Method
+ if method == "" {
+ method = "POST"
+ }
+
+ // Create HTTP request
+ req, err := http.NewRequest(method, webhookConfig.URL, bytes.NewBuffer(payload))
+ if err != nil {
+ log.Error().Err(err).Str("job_id", job.ID).Str("webhook_url", webhookConfig.URL).Msg("Failed to create webhook request")
+ return fmt.Errorf("failed to create request: %w", err)
+ }
+
+ // Set headers
+ req.Header.Set("Content-Type", "application/json")
+ for key, value := range webhookConfig.Headers {
+ req.Header.Set(key, value)
+ }
+
+ // Execute with retry
+ client := &http.Client{Timeout: 30 * time.Second}
+ err = s.executeWithRetry(client, req)
+ if err != nil {
+ log.Error().Err(err).Str("job_id", job.ID).Str("webhook_url", webhookConfig.URL).Msg("Webhook delivery failed")
+ return fmt.Errorf("webhook delivery failed: %w", err)
+ }
+
+ log.Info().Str("job_id", job.ID).Str("webhook_url", webhookConfig.URL).Msg("Webhook delivered successfully")
+ return nil
+}
+
+// buildWebhookPayload builds webhook payload (default or template)
+func (s *AgentJobService) buildWebhookPayload(job schema.Job, task schema.Task, webhookConfig schema.WebhookConfig) ([]byte, error) {
+ if webhookConfig.PayloadTemplate != "" {
+ // Use custom template
+ return s.buildPayloadFromTemplate(job, task, webhookConfig.PayloadTemplate)
+ }
+
+ // Use default format
+ // Include Error field (empty string if no error)
+ payload := map[string]interface{}{
+ "job_id": job.ID,
+ "task_id": job.TaskID,
+ "task_name": task.Name,
+ "status": string(job.Status),
+ "result": job.Result,
+ "error": job.Error, // Empty string if no error
+ "parameters": job.Parameters,
+ "started_at": job.StartedAt,
+ "completed_at": job.CompletedAt,
+ }
+
+ return json.Marshal(payload)
+}
+
+// buildPayloadFromTemplate builds payload from template
+func (s *AgentJobService) buildPayloadFromTemplate(job schema.Job, task schema.Task, templateStr string) ([]byte, error) {
+ // Create template context
+ // Available variables:
+ // - .Job - Job object with all fields
+ // - .Task - Task object
+ // - .Result - Job result (if successful)
+ // - .Error - Error message (if failed, empty string if successful)
+ // - .Status - Job status string
+ ctx := map[string]interface{}{
+ "Job": job,
+ "Task": task,
+ "Result": job.Result,
+ "Error": job.Error,
+ "Parameters": job.Parameters,
+ "Status": string(job.Status),
+ }
+
+ // Add json function for template
+ funcMap := template.FuncMap{
+ "json": func(v interface{}) string {
+ b, _ := json.Marshal(v)
+ return string(b)
+ },
+ }
+
+ tmpl, err := template.New("payload").Funcs(funcMap).Funcs(sprig.FuncMap()).Parse(templateStr)
+ if err != nil {
+ return nil, err
+ }
+
+ var buf bytes.Buffer
+ if err := tmpl.Execute(&buf, ctx); err != nil {
+ return nil, err
+ }
+
+ return buf.Bytes(), nil
+}
+
+// executeWithRetry executes HTTP request with retry logic
+func (s *AgentJobService) executeWithRetry(client *http.Client, req *http.Request) error {
+ maxRetries := 3
+ backoff := []time.Duration{1 * time.Second, 2 * time.Second, 4 * time.Second}
+
+ var err error
+ for i := 0; i < maxRetries; i++ {
+ // Recreate request body if needed (it may have been consumed)
+ if req.Body != nil {
+ bodyBytes, _ := io.ReadAll(req.Body)
+ req.Body.Close()
+ req.Body = io.NopCloser(bytes.NewBuffer(bodyBytes))
+ }
+ var resp *http.Response
+ resp, err = client.Do(req)
+ if err == nil && resp.StatusCode >= 200 && resp.StatusCode < 300 {
+ resp.Body.Close()
+ return nil // Success
+ }
+
+ if resp != nil {
+ resp.Body.Close()
+ }
+
+ if i < maxRetries-1 {
+ time.Sleep(backoff[i])
+ }
+ }
+
+ return fmt.Errorf("failed after %d retries: %w", maxRetries, err)
+}
+
+// CleanupOldJobs removes jobs older than retention period
+func (s *AgentJobService) CleanupOldJobs() error {
+ cutoff := time.Now().AddDate(0, 0, -s.retentionDays)
+ allJobs := s.jobs.Values()
+ removed := 0
+
+ for _, job := range allJobs {
+ if job.CreatedAt.Before(cutoff) {
+ s.jobs.Delete(job.ID)
+ removed++
+ }
+ }
+
+ if removed > 0 {
+ log.Info().Int("removed", removed).Int("retention_days", s.retentionDays).Msg("Cleaned up old jobs")
+ // Save to file
+ if err := s.SaveJobsToFile(); err != nil {
+ log.Error().Err(err).Msg("Failed to save jobs to file after cleanup")
+ }
+ }
+
+ return nil
+}
+
+// Start starts the background service
+func (s *AgentJobService) Start(ctx context.Context) error {
+ // Create service context
+ s.ctx, s.cancel = context.WithCancel(ctx)
+
+ // Update retention days from config
+ retentionDays := s.appConfig.AgentJobRetentionDays
+ if retentionDays == 0 {
+ retentionDays = 30 // Default
+ }
+ s.retentionDays = retentionDays
+
+ // Load tasks and jobs from files
+ if err := s.LoadTasksFromFile(); err != nil {
+ log.Warn().Err(err).Msg("Failed to load tasks from file")
+ }
+ if err := s.LoadJobsFromFile(); err != nil {
+ log.Warn().Err(err).Msg("Failed to load jobs from file")
+ }
+
+ // Start cron scheduler
+ s.cronScheduler.Start()
+
+ // Start worker pool (5 workers)
+ workerCount := 5
+ for i := 0; i < workerCount; i++ {
+ go s.worker(s.ctx)
+ }
+
+ // Schedule daily cleanup at midnight
+ _, err := s.cronScheduler.AddFunc("0 0 * * *", func() {
+ if err := s.CleanupOldJobs(); err != nil {
+ log.Error().Err(err).Msg("Failed to cleanup old jobs")
+ }
+ })
+ if err != nil {
+ log.Warn().Err(err).Msg("Failed to schedule daily cleanup")
+ }
+
+ // Run initial cleanup
+ if err := s.CleanupOldJobs(); err != nil {
+ log.Warn().Err(err).Msg("Failed to run initial cleanup")
+ }
+
+ log.Info().Int("retention_days", s.retentionDays).Msg("AgentJobService started")
+ return nil
+}
+
+// Stop stops the agent job service
+func (s *AgentJobService) Stop() error {
+ if s.cancel != nil {
+ s.cancel()
+ s.cancel = nil
+ }
+ if s.cronScheduler != nil {
+ s.cronScheduler.Stop()
+ }
+ log.Info().Msg("AgentJobService stopped")
+ return nil
+}
+
+// UpdateRetentionDays updates the retention days setting
+func (s *AgentJobService) UpdateRetentionDays(days int) {
+ s.retentionDays = days
+ if days == 0 {
+ s.retentionDays = 30 // Default
+ }
+ log.Info().Int("retention_days", s.retentionDays).Msg("Updated agent job retention days")
+}
diff --git a/core/services/agent_jobs_test.go b/core/services/agent_jobs_test.go
new file mode 100644
index 000000000..9851ba1e9
--- /dev/null
+++ b/core/services/agent_jobs_test.go
@@ -0,0 +1,332 @@
+package services_test
+
+import (
+ "context"
+ "os"
+ "time"
+
+ "github.com/mudler/LocalAI/core/config"
+ "github.com/mudler/LocalAI/core/schema"
+ "github.com/mudler/LocalAI/core/services"
+ "github.com/mudler/LocalAI/core/templates"
+ "github.com/mudler/LocalAI/pkg/model"
+ "github.com/mudler/LocalAI/pkg/system"
+
+ . "github.com/onsi/ginkgo/v2"
+ . "github.com/onsi/gomega"
+)
+
+var _ = Describe("AgentJobService", func() {
+ var (
+ service *services.AgentJobService
+ tempDir string
+ appConfig *config.ApplicationConfig
+ modelLoader *model.ModelLoader
+ configLoader *config.ModelConfigLoader
+ evaluator *templates.Evaluator
+ )
+
+ BeforeEach(func() {
+ var err error
+ tempDir, err = os.MkdirTemp("", "agent_jobs_test")
+ Expect(err).NotTo(HaveOccurred())
+
+ systemState := &system.SystemState{}
+ systemState.Model.ModelsPath = tempDir
+
+ appConfig = config.NewApplicationConfig(
+ config.WithDynamicConfigDir(tempDir),
+ config.WithContext(context.Background()),
+ )
+ appConfig.SystemState = systemState
+ appConfig.APIAddress = "127.0.0.1:8080"
+ appConfig.AgentJobRetentionDays = 30
+
+ modelLoader = model.NewModelLoader(systemState, false)
+ configLoader = config.NewModelConfigLoader(tempDir)
+ evaluator = templates.NewEvaluator(tempDir)
+
+ service = services.NewAgentJobService(
+ appConfig,
+ modelLoader,
+ configLoader,
+ evaluator,
+ )
+ })
+
+ AfterEach(func() {
+ os.RemoveAll(tempDir)
+ })
+
+ Describe("Task CRUD operations", func() {
+ It("should create a task", func() {
+ task := schema.Task{
+ Name: "Test Task",
+ Description: "Test Description",
+ Model: "test-model",
+ Prompt: "Hello {{.name}}",
+ Enabled: true,
+ }
+
+ id, err := service.CreateTask(task)
+ Expect(err).NotTo(HaveOccurred())
+ Expect(id).NotTo(BeEmpty())
+
+ retrieved, err := service.GetTask(id)
+ Expect(err).NotTo(HaveOccurred())
+ Expect(retrieved.Name).To(Equal("Test Task"))
+ Expect(retrieved.Description).To(Equal("Test Description"))
+ Expect(retrieved.Model).To(Equal("test-model"))
+ Expect(retrieved.Prompt).To(Equal("Hello {{.name}}"))
+ })
+
+ It("should update a task", func() {
+ task := schema.Task{
+ Name: "Original Task",
+ Model: "test-model",
+ Prompt: "Original prompt",
+ }
+
+ id, err := service.CreateTask(task)
+ Expect(err).NotTo(HaveOccurred())
+
+ updatedTask := schema.Task{
+ Name: "Updated Task",
+ Model: "test-model",
+ Prompt: "Updated prompt",
+ }
+
+ err = service.UpdateTask(id, updatedTask)
+ Expect(err).NotTo(HaveOccurred())
+
+ retrieved, err := service.GetTask(id)
+ Expect(err).NotTo(HaveOccurred())
+ Expect(retrieved.Name).To(Equal("Updated Task"))
+ Expect(retrieved.Prompt).To(Equal("Updated prompt"))
+ })
+
+ It("should delete a task", func() {
+ task := schema.Task{
+ Name: "Task to Delete",
+ Model: "test-model",
+ Prompt: "Prompt",
+ }
+
+ id, err := service.CreateTask(task)
+ Expect(err).NotTo(HaveOccurred())
+
+ err = service.DeleteTask(id)
+ Expect(err).NotTo(HaveOccurred())
+
+ _, err = service.GetTask(id)
+ Expect(err).To(HaveOccurred())
+ })
+
+ It("should list all tasks", func() {
+ task1 := schema.Task{Name: "Task 1", Model: "test-model", Prompt: "Prompt 1"}
+ task2 := schema.Task{Name: "Task 2", Model: "test-model", Prompt: "Prompt 2"}
+
+ _, err := service.CreateTask(task1)
+ Expect(err).NotTo(HaveOccurred())
+ _, err = service.CreateTask(task2)
+ Expect(err).NotTo(HaveOccurred())
+
+ tasks := service.ListTasks()
+ Expect(len(tasks)).To(BeNumerically(">=", 2))
+ })
+ })
+
+ Describe("Job operations", func() {
+ var taskID string
+
+ BeforeEach(func() {
+ task := schema.Task{
+ Name: "Test Task",
+ Model: "test-model",
+ Prompt: "Hello {{.name}}",
+ Enabled: true,
+ }
+ var err error
+ taskID, err = service.CreateTask(task)
+ Expect(err).NotTo(HaveOccurred())
+ })
+
+ It("should create and queue a job", func() {
+ params := map[string]string{"name": "World"}
+ jobID, err := service.ExecuteJob(taskID, params, "test")
+ Expect(err).NotTo(HaveOccurred())
+ Expect(jobID).NotTo(BeEmpty())
+
+ job, err := service.GetJob(jobID)
+ Expect(err).NotTo(HaveOccurred())
+ Expect(job.TaskID).To(Equal(taskID))
+ Expect(job.Status).To(Equal(schema.JobStatusPending))
+ Expect(job.Parameters).To(Equal(params))
+ })
+
+ It("should list jobs with filters", func() {
+ params := map[string]string{}
+ jobID1, err := service.ExecuteJob(taskID, params, "test")
+ Expect(err).NotTo(HaveOccurred())
+
+ time.Sleep(10 * time.Millisecond) // Ensure different timestamps
+
+ jobID2, err := service.ExecuteJob(taskID, params, "test")
+ Expect(err).NotTo(HaveOccurred())
+
+ allJobs := service.ListJobs(nil, nil, 0)
+ Expect(len(allJobs)).To(BeNumerically(">=", 2))
+
+ filteredJobs := service.ListJobs(&taskID, nil, 0)
+ Expect(len(filteredJobs)).To(BeNumerically(">=", 2))
+
+ status := schema.JobStatusPending
+ pendingJobs := service.ListJobs(nil, &status, 0)
+ Expect(len(pendingJobs)).To(BeNumerically(">=", 2))
+
+ // Verify both jobs are in the list
+ jobIDs := make(map[string]bool)
+ for _, job := range pendingJobs {
+ jobIDs[job.ID] = true
+ }
+ Expect(jobIDs[jobID1]).To(BeTrue())
+ Expect(jobIDs[jobID2]).To(BeTrue())
+ })
+
+ It("should cancel a pending job", func() {
+ params := map[string]string{}
+ jobID, err := service.ExecuteJob(taskID, params, "test")
+ Expect(err).NotTo(HaveOccurred())
+
+ err = service.CancelJob(jobID)
+ Expect(err).NotTo(HaveOccurred())
+
+ job, err := service.GetJob(jobID)
+ Expect(err).NotTo(HaveOccurred())
+ Expect(job.Status).To(Equal(schema.JobStatusCancelled))
+ })
+
+ It("should delete a job", func() {
+ params := map[string]string{}
+ jobID, err := service.ExecuteJob(taskID, params, "test")
+ Expect(err).NotTo(HaveOccurred())
+
+ err = service.DeleteJob(jobID)
+ Expect(err).NotTo(HaveOccurred())
+
+ _, err = service.GetJob(jobID)
+ Expect(err).To(HaveOccurred())
+ })
+ })
+
+ Describe("File operations", func() {
+ It("should save and load tasks from file", func() {
+ task := schema.Task{
+ Name: "Persistent Task",
+ Model: "test-model",
+ Prompt: "Test prompt",
+ }
+
+ id, err := service.CreateTask(task)
+ Expect(err).NotTo(HaveOccurred())
+
+ // Create a new service instance to test loading
+ newService := services.NewAgentJobService(
+ appConfig,
+ modelLoader,
+ configLoader,
+ evaluator,
+ )
+
+ err = newService.LoadTasksFromFile()
+ Expect(err).NotTo(HaveOccurred())
+
+ retrieved, err := newService.GetTask(id)
+ Expect(err).NotTo(HaveOccurred())
+ Expect(retrieved.Name).To(Equal("Persistent Task"))
+ })
+
+ It("should save and load jobs from file", func() {
+ task := schema.Task{
+ Name: "Test Task",
+ Model: "test-model",
+ Prompt: "Test prompt",
+ Enabled: true,
+ }
+
+ taskID, err := service.CreateTask(task)
+ Expect(err).NotTo(HaveOccurred())
+
+ params := map[string]string{}
+ jobID, err := service.ExecuteJob(taskID, params, "test")
+ Expect(err).NotTo(HaveOccurred())
+
+ service.SaveJobsToFile()
+
+ // Create a new service instance to test loading
+ newService := services.NewAgentJobService(
+ appConfig,
+ modelLoader,
+ configLoader,
+ evaluator,
+ )
+
+ err = newService.LoadJobsFromFile()
+ Expect(err).NotTo(HaveOccurred())
+
+ retrieved, err := newService.GetJob(jobID)
+ Expect(err).NotTo(HaveOccurred())
+ Expect(retrieved.TaskID).To(Equal(taskID))
+ })
+ })
+
+ Describe("Prompt templating", func() {
+ It("should build prompt from template with parameters", func() {
+ task := schema.Task{
+ Name: "Template Task",
+ Model: "test-model",
+ Prompt: "Hello {{.name}}, you are {{.role}}",
+ }
+
+ id, err := service.CreateTask(task)
+ Expect(err).NotTo(HaveOccurred())
+
+ // We can't directly test buildPrompt as it's private, but we can test via ExecuteJob
+ // which uses it internally. However, without a real model, the job will fail.
+ // So we'll just verify the task was created correctly.
+ Expect(id).NotTo(BeEmpty())
+ })
+ })
+
+ Describe("Job cleanup", func() {
+ It("should cleanup old jobs", func() {
+ task := schema.Task{
+ Name: "Test Task",
+ Model: "test-model",
+ Prompt: "Test prompt",
+ Enabled: true,
+ }
+
+ taskID, err := service.CreateTask(task)
+ Expect(err).NotTo(HaveOccurred())
+
+ params := map[string]string{}
+ jobID, err := service.ExecuteJob(taskID, params, "test")
+ Expect(err).NotTo(HaveOccurred())
+
+ // Manually set job creation time to be old
+ job, err := service.GetJob(jobID)
+ Expect(err).NotTo(HaveOccurred())
+
+ // Modify the job's CreatedAt to be 31 days ago
+ oldTime := time.Now().AddDate(0, 0, -31)
+ job.CreatedAt = oldTime
+ // We can't directly modify jobs in the service, so we'll test cleanup differently
+ // by setting retention to 0 and creating a new job
+
+ // Test that cleanup runs without error
+ err = service.CleanupOldJobs()
+ Expect(err).NotTo(HaveOccurred())
+ })
+ })
+})
diff --git a/core/services/services_suite_test.go b/core/services/services_suite_test.go
new file mode 100644
index 000000000..21fbfaef6
--- /dev/null
+++ b/core/services/services_suite_test.go
@@ -0,0 +1,13 @@
+package services_test
+
+import (
+ "testing"
+
+ . "github.com/onsi/ginkgo/v2"
+ . "github.com/onsi/gomega"
+)
+
+func TestServices(t *testing.T) {
+ RegisterFailHandler(Fail)
+ RunSpecs(t, "LocalAI services test")
+}
diff --git a/go.mod b/go.mod
index 8efa07025..e2d513a6c 100644
--- a/go.mod
+++ b/go.mod
@@ -62,6 +62,7 @@ require (
require (
github.com/ghodss/yaml v1.0.0 // indirect
github.com/labstack/gommon v0.4.2 // indirect
+ github.com/robfig/cron/v3 v3.0.1 // indirect
github.com/stretchr/testify v1.11.1 // indirect
github.com/swaggo/files/v2 v2.0.2 // indirect
github.com/valyala/fasttemplate v1.2.2 // indirect
diff --git a/go.sum b/go.sum
index 3f25c6a03..6bc71285e 100644
--- a/go.sum
+++ b/go.sum
@@ -666,6 +666,8 @@ github.com/rivo/uniseg v0.1.0/go.mod h1:J6wj4VEh+S6ZtnVlnTBMWIodfgj8LQOQFoIToxlJ
github.com/rivo/uniseg v0.2.0/go.mod h1:J6wj4VEh+S6ZtnVlnTBMWIodfgj8LQOQFoIToxlJtxc=
github.com/rivo/uniseg v0.4.7 h1:WUdvkW8uEhrYfLC4ZzdpI2ztxP1I582+49Oc5Mq64VQ=
github.com/rivo/uniseg v0.4.7/go.mod h1:FN3SvrM+Zdj16jyLfmOkMNblXMcoc8DfTHruCPUcx88=
+github.com/robfig/cron/v3 v3.0.1 h1:WdRxkvbJztn8LMz/QEvLN5sBU+xKpSqwwUO1Pjr4qDs=
+github.com/robfig/cron/v3 v3.0.1/go.mod h1:eQICP3HwyT7UooqI/z+Ov+PtYAWygg1TEWWzGIFLtro=
github.com/rogpeppe/go-internal v1.3.0/go.mod h1:M8bDsm7K2OlrFYOpmOWEs/qY81heoFRclV5y23lUDJ4=
github.com/rogpeppe/go-internal v1.14.1 h1:UQB4HGPB6osV0SQTLymcB4TgvyWu6ZyliaW0tI/otEQ=
github.com/rogpeppe/go-internal v1.14.1/go.mod h1:MaRKkUm5W0goXpeCfT7UZI6fk/L7L7so1lCWt35ZSgc=
diff --git a/core/http/static/chat.js b/core/http/static/chat.js
index 0759b90d7..dc0aef87a 100644
--- a/core/http/static/chat.js
+++ b/core/http/static/chat.js
@@ -2392,7 +2392,10 @@ document.addEventListener('DOMContentLoaded', function() {
if (shouldCreateNewChat) {
// Create a new chat with the model from URL (which matches the selected model from index)
const currentModel = document.getElementById("chat-model")?.value || "";
- const newChat = chatStore.createChat(currentModel, "", indexChatData.mcpMode || false);
+ // Check URL parameter for MCP mode (takes precedence over localStorage)
+ const urlParams = new URLSearchParams(window.location.search);
+ const mcpFromUrl = urlParams.get('mcp') === 'true';
+ const newChat = chatStore.createChat(currentModel, "", mcpFromUrl || indexChatData.mcpMode || false);
// Update context size from template if available
const contextSizeInput = document.getElementById("chat-model");
@@ -2442,8 +2445,16 @@ document.addEventListener('DOMContentLoaded', function() {
}
}, 500);
} else {
// No message, but might have mcpMode - clear localStorage
localStorage.removeItem('localai_index_chat_data');
+
+ // If MCP mode was set from URL, ensure it's enabled
+ const urlParams = new URLSearchParams(window.location.search);
+ if (urlParams.get('mcp') === 'true' && newChat) {
+ newChat.mcpMode = true;
+ saveChatsToStorage();
+ updateUIForActiveChat();
+ }
saveChatsToStorage();
updateUIForActiveChat();
}
@@ -2452,12 +2463,25 @@ document.addEventListener('DOMContentLoaded', function() {
if (!storedData || !storedData.chats || storedData.chats.length === 0) {
const currentModel = document.getElementById("chat-model")?.value || "";
const oldSystemPrompt = localStorage.getItem(SYSTEM_PROMPT_STORAGE_KEY);
- chatStore.createChat(currentModel, oldSystemPrompt || "", false);
+ // Check URL parameter for MCP mode
+ const urlParams = new URLSearchParams(window.location.search);
+ const mcpFromUrl = urlParams.get('mcp') === 'true';
+ chatStore.createChat(currentModel, oldSystemPrompt || "", mcpFromUrl);
// Remove old system prompt key after migration
if (oldSystemPrompt) {
localStorage.removeItem(SYSTEM_PROMPT_STORAGE_KEY);
}
+ } else {
+ // Existing chats loaded - check URL parameter for MCP mode
+ const urlParams = new URLSearchParams(window.location.search);
+ if (urlParams.get('mcp') === 'true') {
+ const activeChat = chatStore.activeChat();
+ if (activeChat) {
+ activeChat.mcpMode = true;
+ saveChatsToStorage();
+ }
+ }
}
// Update context size from template if available
@@ -2442,8 +2445,16 @@ document.addEventListener('DOMContentLoaded', function() {
}, 500);
} else {
- // No message, but might have mcpMode - clear localStorage
+ // No message, but might have mcpMode from URL - clear localStorage
localStorage.removeItem('localai_index_chat_data');